//*****************************************************************************
//* =====================
//*  ENTBSpring
//* =====================
//*  
//* (Description)
//*    e+e- --> entb generator
//*
//* (Update Record)
//*    2018/06/29  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "ENTBSpring.h"

#include <sstream>
#include <iomanip>
//#define __ZEROWIDTH__
//#define __DEBUG__
//#define __PHASESPACE__

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(ENTBSpring)
ClassImp(ENTBSpringBuf)
ClassImp(ENTBBases)
ClassImp(GENPDTTQuark)

//-----------------------------------------------------------------------------
// ==============================
//  class ENTBSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
ENTBSpring::ENTBSpring(const char      *name,
                       const char      *title,
                             ENTBBases *bases)
         : JSFSpring(name, title, bases)
{
  fEventBuf = new ENTBSpringBuf("ENTBSpringBuf",
                                "ENTBSpring event buffer",
                                this);
  if (!bases) { 
    SetBases(new ENTBBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
ENTBSpring::~ENTBSpring()
{
  //delete fEventBuf;
  //delete GetBases();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
Bool_t ENTBSpring::Initialize()
{
  JSFSpring::Initialize();

  if (fFile->IsWritable()) {
    ENTBBases *bs = static_cast<ENTBBases *>(GetBases());
    TDirectory *last = gDirectory;
    fFile->cd("/conf");
    TList *dlist = gDirectory->GetListOfKeys();
    if (!dlist->FindObject("init")) gDirectory->mkdir("init");
    fFile->cd("/conf/init");
    bs->Write();
    last->cd(); 
    cerr << ">>>>>> ENTBBases written to file" << endl;
  }
  return kTRUE;
}

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ==============================
//  class ENTBSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t ENTBSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  ENTBBases    *bases   = static_cast<ENTBBases  *>(
                          static_cast<ENTBSpring *>(Module())->GetBases());

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  fNparton = 9;

  static ANL4DVector *pv = 0;
  if (!pv) {
    pv = new ANL4DVector [fNparton];
  }
  pv[0] = bases->fP[0]; // e-  / e+
  pv[1] = bases->fP[1]; // neb / ne
  pv[3] = bases->fP[2]; // bb  / b
  pv[5] = bases->fP[3]; // b   / bb from t/tb
  pv[7] = bases->fP[4]; // fu  / fub
  pv[8] = bases->fP[5]; // fdb / fd
  pv[6] = pv[7] + pv[8]; // W+ / W- from t/tb
  pv[4] = pv[5] + pv[6]; // t  / tb
  pv[2] = pv[3] + pv[4]; // (t bb)  / (tb b)

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------
  fCP           = bases->GetCP();
  for (Int_t i=0; i<fNparton; i++) {
    pv[i].SetXYZT(fCP*pv[i].X(), fCP*pv[i].Y(), fCP*pv[i].Z(), pv[i].T());
  }
  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fXi           = bases->GetXi();
  fEta0         = bases->GetEta0();
  fEta1         = bases->GetEta1();
  fPhi0         = bases->GetPhi0();
  fPhi1         = bases->GetPhi1();
  fQ2TB         = bases->GetQ2TB();
  fQ2T          = bases->GetQ2T ();
  fQ2W          = bases->GetQ2W ();
  fCosT         = bases->GetCosThetaT();
  fPhiT         = bases->GetPhiT();
  fCosW         = bases->GetCosThetaW();
  fPhiW         = bases->GetPhiW();
  fCosF         = bases->GetCosThetaF();
  fPhiF         = bases->GetPhiF();
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);

  for (Int_t i=0; i<fNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------
  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [fNparton];
    for (Int_t i=0; i<fNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<fNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    idw     = fCP*24;                          // PDG code for W+
  Int_t    idt     = fCP*6;                           // PDG code for t
  Int_t    idbb    = fCP*(-5);                        // PDG code for bb
  Int_t    idb     = fCP*5;                           // PDG code for b
  Int_t    ide     = fCP*11;                          // PDG code for e-
  Int_t    idn     = fCP*(-12);                       // PDG code for neb
  Double_t mn      = 0.;                              // electron mass
  Double_t me      = kM_e;                            // electron mass
  Double_t mb      = bases->fM[3];                    // b mass
  Double_t chrgw   = fCP;
  Double_t chrge   = -fCP;                            // electron charge
  Double_t chrgb   = fCP*(-1./3);                     // b  charge
  Double_t chrgt   = fCP*( 2./3);                     // b  charge
  Int_t    hel2    = fCP*bases->fHelFinal[2];         // bb helicity
  Int_t    hel3    = fCP*bases->fHelFinal[3];         // bb helicity
  Int_t    icb     = 1;                               // color flux id for the b bb system
  Double_t rq2vw   = pv[2].Mag();
  Double_t rq2t    = pv[4].Mag();
  Int_t    islvw   = 101;      	  	 	      // shower level
  Int_t    islvt   = 201;      	  	 	      // shower level

  Int_t    idf4    = fCP*bases->f4Ptr->GetPID   ();   // PDG code for fu
  Double_t chrg4   = fCP*bases->f4Ptr->GetCharge();   // fu charge
  Double_t m4      =     bases->f4Ptr->GetMass  ();   // fu mass
  Int_t    hel4    = fCP*bases->fHelFinal[4];         // fu helicity
  Int_t    idf5    = fCP*bases->f5Ptr->GetPID   ();   // PDG code for fdb
  Double_t chrg5   = fCP*bases->f5Ptr->GetCharge();   // fdb charge
  Double_t m5      =     bases->f5Ptr->GetMass  ();   // fdb mass
  Int_t    hel5    = fCP*bases->fHelFinal[5];         // fdb helicity
  Double_t color   = bases->f4Ptr->GetColor();    // color factor
  Int_t    islev   = color > 1. ? 301 : 0;  	  // shower level
  Int_t    icf     = 2;                           // color flux id
  Double_t rq2w    = pv[6].Mag();
#ifdef __DEBUG__
  cerr << endl;
  ANL4DVector qcm;
  for (Int_t ip=0; ip<9; ip++) {
    cerr << "pv[" << ip << "] = (" 
         <<  pv[ip].E() << ", "
         <<  pv[ip].Px() << ", "
         <<  pv[ip].Py() << ", "
         <<  pv[ip].Pz() << ") "
	 << "m = " << pv[ip].GetMass() << endl;
    if (ip<3) qcm += pv[ip];
  }
  cerr << "qcm = ("
       <<  qcm.E() << ", "
       <<  qcm.Px() << ", "
       <<  qcm.Py() << ", "
       <<  qcm.Pz() << ") " << endl;
  cerr << "----" << endl;
#endif

  //                               No. PID   Mass  Charge   pv   Nd 1st Mom hel  col shower
  new (partons[0]) JSFSpringParton(1, ide ,   me, chrge, *qp[0], 0, 0,  0,    0,   0,     0);
  new (partons[1]) JSFSpringParton(2, idn ,   mn,    0., *qp[1], 0, 0,  0,    0,   0,     0);
  new (partons[2]) JSFSpringParton(3, idw ,rq2vw, chrgw, *qp[2], 2, 4,  0,    0,   0,     0);
  new (partons[3]) JSFSpringParton(4, idbb,   mb,-chrgb, *qp[3], 0, 0,  3, hel2, icb, islvw);
  new (partons[4]) JSFSpringParton(5, idt , rq2t, chrgt, *qp[4], 2, 6,  3,    0,   0, islvw);
  new (partons[5]) JSFSpringParton(6, idb ,   mb, chrgb, *qp[5], 0, 0,  5, hel3, icb, islvt);
  new (partons[6]) JSFSpringParton(7, idw , rq2w, chrgw, *qp[6], 2, 8,  5,    0,   0, islvt);
  new (partons[7]) JSFSpringParton(8, idf4,   m4, chrg4, *qp[7], 0, 0,  7, hel4, icf, islev);
  new (partons[8]) JSFSpringParton(9, idf5,   m5, chrg5, *qp[8], 0, 0,  7, hel5, icf, islev);

  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class ENTBBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
ENTBBases::ENTBBases(const char *name, const char *title)
         : JSFBases   (name, title), 
           fEcmInit   (1000.),
           fTopMass   (175.),
           fISR       ( 1),
           fBeamStr   ( 1),
           fBeamWidth (0.002),
           fPolem     (0.),
           fPolep     (0.),
           fWModesLo  ( 1),
           fWModesHi  (12),
	   fNCALL     (80000),
	   fACC1      (0.05),
	   fACC2      (0.05),
	   fITMX1     (20),
	   fITMX2     (40),
           fWBosonPtr ( 0),
           fZBosonPtr ( 0),
           fTopPtr    ( 0),
           fZBoost    (0.),
           fEcmIP     (fEcmInit),
	   fXi        (0.),
	   fEta0      (0.),
	   fEta1      (0.),
	   fPhi0      (0.),
	   fPhi1      (0.),
	   fQ2TB      (0.),
	   fQ2T       (0.),
	   fQ2W       (0.),
           fWModePtr  (0),
           f4Ptr      (0),
           f5Ptr      (0),
	   fCosT      (0.),
	   fPhiT      (0.),
	   fCosW      (0.),
	   fPhiW      (0.),
	   fCosF      (0.),
	   fPhiF      (0.),
           fCP        ( 1),
	   fSh0       (0.),
	   fCh0       (0.),
           fR_BW_m    (0),
           fR_BW_p    (0),
           fR_BS_m    (0),
           fR_BS_p    (0),
           fR_ISR_var (0),
           fR_ISR_side(0)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init entbbases " << endl;
  
  using namespace std;

  stringstream ins(gJSF->Env()->GetValue("ENTBBases.Ecm","500.")); // E_cm [GeV]
  ins >> fEcmInit;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.TopMass","175."));
  ins >> fTopMass;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.CosthTRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.PhiTOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.CosthWRange","-1.0 1.0"));
  ins >> fXL[2] >> fXU[2];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.PhiWOverPiRange","0.0 2.0"));
  ins >> fXL[3] >> fXU[3];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.CosthFRange","-1.0 1.0"));
  ins >> fXL[4] >> fXU[4];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.PhiFOverPiRange","0.0 2.0"));
  ins >> fXL[5] >> fXU[5];

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.Beamstrahlung","1")); // BmStr (on)
  ins >> fBeamStr;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.BeamWidth","0.002")); // BmStr (on)
  ins >> fBeamWidth;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.Bremsstrahlung","1"));// ISR (on)
  ins >> fISR;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.Polem","0."));       // electron polarization
  ins >> fPolem;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.Polep","0."));       // positron polarization
  ins >> fPolep;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.WModesLo","1"));      // Z decay mode lo
  ins >> fWModesLo;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.WModesHi","12"));     // Z decay mode hi
  ins >> fWModesHi;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.ACC1","0.05"));
  ins >> fACC1;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.ACC2","0.05"));
  ins >> fACC2;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.ITMX1","20"));
  ins >> fITMX1;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.ITMX2","40"));
  ins >> fITMX2;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("ENTBBases.NCALL","80000"));
  ins >> fNCALL;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fXCP           , 0., 1., 1, 1);
  DefineVariable(fHelCombInitial, 0., 1., 1, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 1, 1);
  DefineVariable(fWDecayMode    , 0., 1., 0, 1);
  DefineVariable(fXQ2TB         , 0., 1., 1, 1);
  DefineVariable(fXQ2T          , 0., 1., 1, 1);
  DefineVariable(fXQ2W          , 0., 1., 1, 1);

  //--
  //  xi, eta0, eta1, phi0, phi10:=phi1-phi0
  //--
  DefineVariable(fXXi   , 0., 1., 1, 1);
  DefineVariable(fXEta0 , 0., 1., 1, 1);
  DefineVariable(fXEta1 , 0., 1., 1, 1);
  DefineVariable(fXPhi0 , 0., 1., 1, 1);
  DefineVariable(fXPhi10, 0., 1., 1, 1);

  //--
  //  cos(theta) and phi
  //--
  fXL[1] = fXL[1]*TMath::Pi();
  fXU[1] = fXU[1]*TMath::Pi();
  fXL[3] = fXL[3]*TMath::Pi();
  fXU[3] = fXU[3]*TMath::Pi();
  fXL[5] = fXL[5]*TMath::Pi();
  fXU[5] = fXU[5]*TMath::Pi();

  DefineVariable(fCosT , fXL[0], fXU[0], 0, 1);
  DefineVariable(fPhiT , fXL[1], fXU[1], 0, 1);
  DefineVariable(fCosW , fXL[2], fXU[2], 0, 1);
  DefineVariable(fPhiW , fXL[3], fXU[3], 0, 1);
  DefineVariable(fCosF , fXL[4], fXU[4], 0, 1);
  DefineVariable(fPhiF , fXL[5], fXU[5], 0, 1);

  //--
  //  Beamstrahlung and natural Ebeam spread
  //--
  if (fBeamStr == 1) {
    DefineVariable(fR_BW_m, 0., 1., 0, 1);
    DefineVariable(fR_BW_p, 0., 1., 0, 1);
    DefineVariable(fR_BS_m, 0., 1., 0, 1);
    DefineVariable(fR_BS_p, 0., 1., 0, 1);
  }
 
  //--
  //  ISR
  //--
  if (fISR==1) {
    DefineVariable(fR_ISR_var,  0., 1., 0, 1);
    DefineVariable(fR_ISR_side, 0., 1., 0, 1);
  }

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(fNCALL);

  SetTuneValue (1.5);
  SetIteration1(fACC1, fITMX1);
  SetIteration2(fACC2, fITMX2);

}
// --------------------------
//  D-tor
// --------------------------
ENTBBases::~ENTBBases()
{
  delete fWBosonPtr;
  delete fZBosonPtr;
  delete fTopPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t ENTBBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.; // Jacobian factor
  Double_t eminus;        // E_e- after ISR and beamstrahlung
  Double_t eplus;         // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  if (fBeamStr == 1) {
    bsWeight = fBM->GetWeight(fR_BW_m, fR_BW_p,
                              fR_BS_m, fR_BS_p,
                              eminus, eplus);
    eplus   *= fEcmInit/2.;
    eminus  *= fEcmInit/2.;
    fEcmIP   = 2.*TMath::Sqrt(eplus*eminus);     // ignore electron mass 
  } else {
    fEcmIP   = fEcmInit;      
    eplus    = fEcmInit/2.;
    eminus   = eplus;
  }
   
  // --------------------------------------------
  //  Initial State Radiation
  // --------------------------------------------

  static const Double_t kFisr = 2. * kAlpha0 / TMath::Pi();

  Double_t beta_isr = kFisr * (2.*(TMath::Log(fEcmIP/kM_e) - 1.));

  Double_t qed = 1.;
  if (fISR == 1) {
    qed     = (1. + 3.*beta_isr/4.)
              *(1. + kFisr * ( TMath::Pi()*TMath::Pi()/6. - 1./4. ));
    Double_t xisr    = TMath::Power(fR_ISR_var,1./beta_isr); // Ephoton / Ebeam 
             fEcmIP *= TMath::Sqrt(1. - xisr);               // reduced Ecm
    if (fR_ISR_side > 0.5) {
      eplus  *= 1. - xisr;
    } else {
      eminus *= 1. - xisr;
    }
  } // if Bremsstrahlung


  fZBoost = eminus - eplus; // P_z of the cm system after ISR and beamstrahlung

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Int_t    igb  = 2;
  Int_t    itb  = 1;
  Int_t    lqb  = 1;
  Double_t m2   = kMass[lqb][itb][igb]; // bb
  Double_t m3   = kMass[lqb][itb][igb]; // b from t

  Double_t weight = 1;
  GENDecayMode *fWModePtr = fWBosonPtr->PickMode(fWDecayMode, weight, fWMode);
  bsWeight *= weight;
  f4Ptr = static_cast<GENPDTEntry *>(fWModePtr->At(0));
  f5Ptr = static_cast<GENPDTEntry *>(fWModePtr->At(1));
  Double_t m4   = f4Ptr->GetMass();
  Double_t m5   = f5Ptr->GetMass();

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  Double_t m0   = kM_e; // Me
  Double_t m1   = 0.;   // Mneb
  if (fEcmIP < m0 + m1 + m2 + m3 + m4 + m5) {
    return 0.;
  }

  // --------------------------------------------
  //  Flip CP
  // --------------------------------------------
  if (fXCP <= 0.5) fCP = -1;
  else             fCP = +1;
  bsWeight *= 2.;  

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average is taken care of here
  SelectHelicities(weight);
  if (weight == 0.) return 0.;
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------

  Double_t rs   = fEcmIP;
  Double_t qmin = m4 + m5;
  Double_t qmax = rs - (m0 + m1 + m2 + m3);
#ifndef __ZEROWIDTH__
  fQ2W = fWBosonPtr->GetQ2BW(qmin, qmax, fXQ2W, weight);
#else
  fQ2W = TMath::Power(fWBosonPtr->GetMass(),2);
  weight = kPi*fWBosonPtr->GetMass()*fWBosonPtr->GetWidth();
#endif
  bsWeight *= weight;

  Double_t qw   = TMath::Sqrt(fQ2W);       // Mw
#ifndef __ZEROWIDTH__
  qmin = m3 + qw;
  qmax = rs - (m0 + m1 + m2);
  fQ2T = fTopPtr->GetQ2BW(qmin, qmax, fXQ2T, weight);
#else
  fQ2T = TMath::Power(fTopPtr->GetMass(),2);
  qmin = TMath::Sqrt(fQ2T);
  qmax = rs - (m0 + m1 + m2);
  if (qmin >= qmax) return 0.;
  weight = kPi*fTopPtr->GetMass()*fTopPtr->GetWidth();
#endif
  bsWeight *= weight;

  Double_t qt   = TMath::Sqrt(fQ2T);       // Mt
  qmin = m2 + qt;
  qmax = rs - (m0 + m1);
  Double_t q2min = qmin*qmin;
  Double_t q2max = qmax*qmax;
  fQ2TB  = q2min + fXQ2TB*(q2max-q2min);
  weight = q2max - q2min;
  bsWeight *= weight;

  // --------------------------------------------
  //  Handle kinematics here
  // --------------------------------------------
  Double_t s    = rs*rs;
  Double_t qtb  = TMath::Sqrt(fQ2TB);

  Double_t xilo = TMath::Log((qtb-m0)*(qtb+m0)/s);
  Double_t xihi = TMath::Log(1.-2.*m0/rs);
           xilo = TMath::Min(xihi,xilo);
  fXi = xilo + (xihi-xilo)*fXXi;
  bsWeight *= xihi - xilo;
  Double_t rxi  = TMath::Exp(fXi);

  Double_t dme  = (m0*m0/s)*rxi*rxi/(1.-rxi);
  Double_t dpe  = 1.;

  Double_t etlo = -TMath::Log((1.+dme)/dpe)/2.;
  Double_t ethi =  TMath::Log((1.+dpe)/dme)/2.;
  fEta0     = etlo + (ethi-etlo)*fXEta0;
  bsWeight *= ethi - etlo;

  fEta1  = fXEta1;

  fPhi0  = k2Pi * fXPhi0;
  fPhi1  = k2Pi * fXPhi10 + fPhi0;
  fPhi1  = fPhi1 > k2Pi ? fPhi1 - k2Pi : fPhi1;
  bsWeight *= k2Pi*k2Pi;

  fM[0] = m0;
  fM[1] = m1;
  fM[2] = m2;
  fM[3] = m3;
  fM[4] = m4;
  fM[5] = m5;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------

  Double_t sigma = DSigmaDX() * qed;

#if 1
  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------
  Double_t qwn2 = s*rxi + m0*m0;
  Double_t qwn  = TMath::Sqrt(qwn2);
  Double_t exp2et0 = TMath::Exp(2.*fEta0);
  Double_t sh0     = TMath::Min(TMath::Sqrt((1.+dpe+dme)/(1.+exp2et0) - dme), 1.);
  Double_t mw    = fWBosonPtr->GetMass();
  Double_t dlt   = 2.*mw*mw/((qwn-mw)*(qwn+mw));
  Double_t opdzt = TMath::Power(1. + 1./dlt, fEta1);
  Double_t sh1   = TMath::Min(TMath::Sqrt(dlt*(opdzt - 1.)), 1.);
  Double_t xe    = 1. + m0*m0/s - qwn2/s;
  const Double_t kSqrt2 = TMath::Sqrt(2.);
  Double_t cose  = (1. - kSqrt2*sh0)*(1. + kSqrt2*sh0);
  Double_t cosn  = (1. - kSqrt2*sh1)*(1. + kSqrt2*sh1);
  ANL4DVector ptee  = fP[0] - fK[0];
  Double_t    aptee = ptee.P();
  Double_t    aetee = ptee.E();
  Double_t    tee   = (aetee-aptee)*(aetee+aptee);
  ANL4DVector pvw   = fP[2] + fP[3] + fP[4] + fP[5];
  Double_t    costb = pvw.CosTheta();
  Double_t    phitb = pvw.Phi();
  Double_t    xtw   = 2.*pvw.Pt()/rs;
  Double_t    xel   = 2.*fP[0].E()/rs;
  Double_t    xnu   = 2.*fP[1].E()/rs;
  Double_t    xwp   = 2.*pvw.E()/rs;

  Xh_fill( 1, fEcmIP           , (bsWeight*sigma));
  Xh_fill( 2, fXi              , (bsWeight*sigma));
  Xh_fill( 3, fEta0            , (bsWeight*sigma));
  Xh_fill( 4, fEta1            , (bsWeight*sigma));
  Xh_fill( 5, fPhi0            , (bsWeight*sigma));
  Xh_fill( 6, fPhi1            , (bsWeight*sigma));
  Xh_fill( 7, k2Pi*fXPhi10     , (bsWeight*sigma));
  Xh_fill( 8, qtb              , (bsWeight*sigma));
  Xh_fill( 9, costb            , (bsWeight*sigma));
  Xh_fill(10, phitb            , (bsWeight*sigma));
  Xh_fill(11, qt               , (bsWeight*sigma));
  Xh_fill(12, fCosT            , (bsWeight*sigma));
  Xh_fill(13, fPhiT            , (bsWeight*sigma));
  Xh_fill(14, qw               , (bsWeight*sigma));
  Xh_fill(15, fCosF            , (bsWeight*sigma));
  Xh_fill(16, fPhiF            , (bsWeight*sigma));
  Xh_fill(17, (Double_t)fJCombI, (bsWeight*sigma));
  Xh_fill(18, (Double_t)fJCombF, (bsWeight*sigma));
  Xh_fill(19, (Double_t)fWMode , (bsWeight*sigma));
  Xh_fill(20, (Double_t)fCP    , (bsWeight*sigma));
  Xh_fill(21, cose             , (bsWeight*sigma));
  Xh_fill(22, xe               , (bsWeight*sigma));
  Xh_fill(23, cosn             , (bsWeight*sigma));
  Xh_fill(24, xtw              , (bsWeight*sigma));
  Xh_fill(25, xel              , (bsWeight*sigma));
  Xh_fill(26, xnu              , (bsWeight*sigma));
  Xh_fill(27, xwp              , (bsWeight*sigma));
  Xh_fill(28, tee              , (bsWeight*sigma));
#endif

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t ENTBBases::DSigmaDX()
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t weight = 1.;

  Double_t rs   = fEcmIP;
  Double_t s    = rs*rs; 
  Double_t eb   = rs/2.;
  Double_t qtb2 = fQ2TB;
  Double_t qtb  = TMath::Sqrt(qtb2); 
  Double_t m0   = fM[0]; // Me
  Double_t m1   = fM[1]; // Mneb

  Double_t rxi   = TMath::Exp(fXi);
  Double_t qtbn2 = s*rxi + m0*m0;
  Double_t qtbn  = TMath::Sqrt(qtbn2);
  weight *= s*rxi;

  Double_t dm0 = (m0*m0/s)*rxi*rxi/(1.-rxi);
  Double_t dp0 = 1.;

  Double_t exp2et0 = TMath::Exp(2.*fEta0);
  Double_t sh0     = TMath::Min(TMath::Sqrt((1.+dp0+dm0)/(1.+exp2et0) - dm0), 1.);
  Double_t ch0     = TMath::Sqrt((1-sh0)*(1+sh0));
  static Double_t kSqrt2 = TMath::Sqrt(2.);
  Double_t cs0     = (1.-kSqrt2*sh0)*(1.+kSqrt2*sh0);
  Double_t sn0     = 2.*sh0*ch0;
  Double_t fi0     = fPhi0;
  weight *= 4.* (1.+dp0+dm0)/((1.+exp2et0)*(1.+1./exp2et0));

  Double_t dlt = 10.;
  if (qtb > 0. && qtbn > qtb) {
    dlt = 2.*qtb2/((qtbn-qtb)*(qtbn+qtb));
  }

  Double_t opdzt = TMath::Power(1. + 1./dlt, fEta1);
  Double_t sh1   = TMath::Min(TMath::Sqrt(dlt*(opdzt - 1.)), 1.);
  Double_t cs1   = (1.-kSqrt2*sh1)*(1.+kSqrt2*sh1);
  Double_t fi1   = fPhi1;
  weight *= 2.*dlt*opdzt*TMath::Log(1.+1./dlt);

  Double_t pb     = TMath::Sqrt((eb-kM_e)*(eb+kM_e));
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, eb);

  GENFrame    cmframe;
  ANL4DVector qcm(rs, 0.,0.,0.);
  Double_t    m02 = m0*m0;
  GENPhase2   phaseCM(qcm, m02, qtbn2, cmframe, cs0, fi0, 0);
              fP[0] = phaseCM.GetFourMomentum(0);
  ANL4DVector ptbn  = phaseCM.GetFourMomentum(1);
  Double_t    betae = phaseCM.GetBetaBar();
  if (betae <= 0.) return 0.;

  Double_t ap1 = fP[0].P();
  fP[0].SetX(ap1*sn0*TMath::Cos(fi0));
  fP[0].SetY(ap1*sn0*TMath::Sin(fi0));

  Double_t    m12 = m1*m1;
  GENPhase2   phaseWN(ptbn, m12, qtb2, cmframe, cs1, fi1, 1);
              fP[1] = phaseWN.GetFourMomentum(0);
  ANL4DVector ptb   = phaseWN.GetFourMomentum(1);
  Double_t    betan = phaseWN.GetBetaBar();
  if (betan <= 0.) return 0.;

  Double_t m2  = fM[2]; // Mbb
  Double_t m22 = m2*m2;
  Double_t qt2 = fQ2T;
  GENFrame wnframe = phaseWN.GetFrame();
  GENPhase2   phaseTB(ptb, qt2, m22, wnframe, fCosT, fPhiT, 1);
              fP[2] = phaseTB.GetFourMomentum(1);
  ANL4DVector pt    = phaseTB.GetFourMomentum(0);
  Double_t betatb = phaseTB.GetBetaBar();
  if (betatb <= 0.) return 0.;

  Double_t m3  = fM[3]; // Mb
  Double_t m32 = m3*m3;
  Double_t qw2 = fQ2W;
  GENFrame tbframe = phaseTB.GetFrame();
  GENPhase2   phaseT(pt, qw2, m32, tbframe, fCosW, fPhiW, 1);
              fP[3] = phaseT.GetFourMomentum(1);
  ANL4DVector pw    = phaseT.GetFourMomentum(0);
  Double_t betab = phaseT.GetBetaBar();
  if (betab <= 0.) return 0.;

  Double_t m4  = fM[4]; // Mfu
  Double_t m5  = fM[5]; // Mfdb
  Double_t m42 = m4*m4;
  Double_t m52 = m5*m5;
  GENFrame tframe = phaseT.GetFrame();
  GENPhase2   phaseW(pw, m42, m52, tframe, fCosF, fPhiF, 1);
              fP[4] = phaseW.GetFourMomentum(0);
              fP[5] = phaseW.GetFourMomentum(1);
  Double_t betaf = phaseW.GetBetaBar();
  if (betaf <= 0.) return 0.;

  fSh0 = sh0;
  fCh0 = ch0;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
#ifdef __DEBUG__
  for (Int_t i=0; i<6; i++) {
    cerr << " fP[" << i << "] = (" 
         << fP[i].E () << ","
         << fP[i].Px() << ","
         << fP[i].Py() << ","
         << fP[i].Pz() << ")" << endl;
  }
  ANL4DVector qvw = fP[4] + fP[5];
  cerr << " pw  = (" 
       << qvw.E () << ","
       << qvw.Px() << ","
       << qvw.Py() << ","
       << qvw.Pz() << ")" << endl;
  ANL4DVector qvt = qvw + fP[3];
  cerr << " pt  = (" 
       << qvt.E () << ","
       << qvt.Px() << ","
       << qvt.Py() << ","
       << qvt.Pz() << ")" << endl;
  ANL4DVector pcm = fP[0] + fP[1] + fP[2] + qvt;
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << pcm.Py() << ","
       << pcm.Pz() << ")" << endl;
  cerr << " wmass = " << qvw.GetMass() << endl;
#endif
  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared();
  
  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbr  = 5;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));

  Double_t identp = 1.;                            // identical particle factor
  Double_t dPhase = kFact * betae * betan * betatb * betab * betaf 
                          * weight;                // phase space factor
  Double_t flux   = 1./(2.* s * beta_e);           // beam flux factor

  Double_t sigma  = identp * flux * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                       // now in [fb]
#if 0
	   cerr << std::scientific
	        << "weight = " << weight << " betae = "  << betae
		<< " betan = " << betan  << " betatb = " << betatb 
		<< " betab = " << betab  << " betaf = "  << betaf 
		<< " identp = " << identp << " flux = " << flux
		<< " kFact = " << kFact << " amp2 = " << amp2
		<< " beta_e = " << beta_e << " s = " << s
		<< " dPhase = " << dPhase << " sigma = " << sigma << endl;
#endif

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t ENTBBases::AmpSquared()
{
#ifndef __PHASESPACE__
  Double_t  qed    = kAlpha0/kAlpha;       // almost real photon dominance
  Double_t  color  = 3*(1. + kAlphaS/kPi); // color factor for (tb)
            color *= f4Ptr->GetColor();    // color factor for W+ -> fu fdb
  Int_t     ig4    = f4Ptr->GetGenNo() - 1;
  Int_t     ig5    = f5Ptr->GetGenNo() - 1;
  Double_t  mix    = TMath::Power(kVkm[ig4][ig5],2);

  Complex_t amp    = FullAmplitude();
  Double_t  amp2   = TMath::Power(abs(amp),2) * qed * color * mix;

  return amp2;
#else
  return 1.;
#endif
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t ENTBBases::FullAmplitude()
{
   Double_t mt     = fTopPtr->GetMass();
   Double_t gamt   = fTopPtr->GetWidth();
   Double_t mw     = fWBosonPtr->GetMass();
   Double_t gamw   = fWBosonPtr->GetWidth();
   Double_t glw    = -kGw*kSqh;
   Double_t grw    = 0.;

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em (fK[0], kM_e,  fHelInitial[0], +1, kIsIncoming);
   HELFermion ep (fK[1], kM_e,  fHelInitial[1], -1, kIsOutgoing);

   HELFermion e  (fP[0], fM[0], fHelFinal  [0], +1, kIsOutgoing);
   HELFermion neb(fP[1], fM[1], fHelFinal  [1], -1, kIsIncoming);

   HELFermion fu (fP[4], fM[4], fHelFinal  [4], +1, kIsOutgoing);
   HELFermion fdb(fP[5], fM[5], fHelFinal  [5], -1, kIsIncoming);
   HELVector  wf(fdb, fu, glw, grw, mw, gamw);

   HELFermion b  (fP[3], fM[3], fHelFinal  [3], +1, kIsOutgoing);
   HELFermion t  (b, wf, glw, grw, mt, gamt);

   HELFermion bb (fP[2], fM[2], fHelFinal  [2], -1, kIsIncoming);

   Complex_t amp = AmpEEtoENTB(em, ep, e, neb, t, bb);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEEtoENTB()
// --------------------------
Complex_t ENTBBases::AmpEEtoENTB(const HELFermion &em,  // e-
                                 const HELFermion &ep,  // e+
                                 const HELFermion &emf, // e-
                                 const HELFermion &nbf, // nu_e-bar
                                 const HELFermion &t,   // t
                                 const HELFermion &bb)  // b-bar
{
   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
   // ENTB Production Amplitude
   //---------------------------
   Double_t  mn    = kMass[0][0][0];
   Double_t  me    = kMass[0][1][0];
   Double_t  gamn  = 0.;
   Double_t  game  = 0.;

   Double_t  mw    = fWBosonPtr->GetMass();
   Double_t  gamw  = fWBosonPtr->GetWidth();
   Double_t  mz    = fZBosonPtr->GetMass();
   Double_t  gamz  = fZBosonPtr->GetWidth();

   Double_t  mt    = fTopPtr->GetMass();
   Double_t  gamt  = fTopPtr->GetWidth();
   Double_t  mb    = kMass[1][1][2];
   Double_t  gamb  = 0.;

   Double_t  glw   = -kGw*kSqh;
   Double_t  grw   = 0.;

   Double_t  qe    = -1.;
   Double_t  glae  = -kGe*qe;
   Double_t  grae  = -kGe*qe;

   Double_t  t3e   = -1./2.;
   Double_t  glze  = -kGz*(t3e - qe*kSin2W);
   Double_t  grze  = -kGz*(    - qe*kSin2W);

   Double_t  t3n   = +1./2.;
   Double_t  glzn  = -kGz*t3n;
   Double_t  grzn  = 0.;

   Double_t  qt    = 2./3;
   Double_t  glat  = -kGe*qt;
   Double_t  grat  = -kGe*qt;

   Double_t  t3t   = +1./2.;
   Double_t  glzt  = -kGz*(t3t - qt*kSin2W);
   Double_t  grzt  = -kGz*(    - qt*kSin2W);

   Double_t  qb    = -1./3;
   Double_t  glab  = -kGe*qb;
   Double_t  grab  = -kGe*qb;

   Double_t  t3b   = -1./2.;
   Double_t  glzb  = -kGz*(t3b - qb*kSin2W);
   Double_t  grzb  = -kGz*(    - qb*kSin2W);

   Double_t  gwwa  = kGe;     
   Double_t  gwwz  = kGw*kCosW;

   HELVector wf(bb, t, glw, grw, mw, gamw);

   Double_t  ebm   = TMath::Abs(em .GetFourMomentum()(0));
   Double_t  e1    = TMath::Abs(emf.GetFourMomentum()(0));

   HELVector  av (ebm, e1, fSh0, fCh0, fPhi0, 
                  fHelInitial[0], fHelFinal[0],+1, kGe, me);  	// gamma* from e-
   HELVector  zv (em , emf, glze, grze, mz, gamz);		// Z* from e-
   HELVector  wv (nbf, ep , glw, grw, mw, gamw);		// W* from e+

   HELFermion nbv (ep , wf, glw, grw, mn, gamn);		// nb* after emitting Wf
   HELFermion epv (nbf, wf, glw, grw, me, game);		// e+* before emitting Wf
   HELFermion nv  (emf, wf, glw, grw, mn, gamn);		// n*  before emitting Wf

   //-------------------------------------
   // ENW* Amplitudes
   //   graph numbers from hep-ph-9401295
   //-------------------------------------
   HELVertex  amp01 (nbf, nbv, zv, glzn, grzn);	// nbf nb* Z*    (graph 1)

   HELVertex  amp02 (em, nv, wv, glw, grw);	// em n* W*      (graph 2)

   HELVertex  amp03 (epv, ep, av, glae, grae); 	// e+* e+ gamma* (graph 3)
   HELVertex  amp04 (epv, ep, zv, glze, grze); 	// e+* e+ Z*     (graph 4)

   HELVertex  amp05 (wv, wf, av, gwwa);		// W* wf gamma*  (graph 5)
   HELVertex  amp06 (wv, wf, zv, gwwz);		// W* wf Z*      (graph 6)

   Complex_t amp1 = amp01 + amp02 + amp03 + amp04 + amp05 + amp06;

   //-------------------------------------
   // t-channel t/b exchange Amplitudes
   //-------------------------------------
   HELFermion tv  (bb, wv, glw, grw, mt, gamt); // t-channel t*
   HELFermion bbv (t , wv, glw, grw, mb, gamb); // t-channel b*

   HELVertex  amp07 (tv, t  , av, glat, grat);	// t* t a*      (graph  7)
   HELVertex  amp08 (tv, t  , zv, glzt, grzt);	// t* t Z*      (graph  8)
   HELVertex  amp09 (bb, bbv, av, glab, grab);	// bb bb* a*    (graph  9)
   HELVertex  amp10 (bb, bbv, zv, glzb, grzb);	// bb bb* Z*    (graph 10)

   Complex_t amp2 = amp07 + amp08 + amp09 + amp10;

   //-------------------------------------
   // All amplitudes
   //-------------------------------------
   Complex_t amp  = amp1 + amp2;
#endif /* end __PHASESPACE__ */

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void ENTBBases::Userin()
{
  // --------------------------------------------
  //  Open beamstrahlung data
  // --------------------------------------------
  string bsfiledir;
  string bsfilename;

  if (fBeamStr) {
    stringstream ins(gJSF->Env()->GetValue("ENTBBases.BeamstrahlungFilepath",
                                  "/proj/soft/jsf/pro/data/bsdata/"));
    ins >> bsfiledir;

    ins.clear();
    ins.str(gJSF->Env()->GetValue("ENTBBases.BeamstrahlungFilename","trc500"));
    ins >> bsfilename;
    
    TFile *fBeamFile = new TFile((bsfiledir+bsfilename+".root").data());
    if (!fBeamFile) {
      cerr << " Unable to open a file for beamstrahlung" << endl;
      abort();
    }
    
  // --------------------------------------------
  //  Initialize beam generator
  // --------------------------------------------

    fBM = (JSFBeamGenerationCain*)fBeamFile->Get(bsfilename.data());
    fBM->SetIBParameters(fBeamWidth);

    fBM->MakeBSMap();
    fBM->Print(); 

    cout << " Nominal Energy (beamstrahlung) = " << fBM->GetNominalEnergy() 
         << " beam width     (beamstrahlung) ="  << fBM->GetIBWidth() 
         << endl;

    if (!(fBM->GetNominalEnergy() == (fEcmInit/2))) {
      cout << "Nominal energy from beamstrahung is " << fBM->GetNominalEnergy() 
           << " which is different from fEcm/2: "    << (fEcmInit/2)
           << endl;        
    } // check the energy homogeneity
  } // if beamstrahlung is on

  // --------------------------------------------
  //  Initialize W decay table
  // --------------------------------------------
  if (! fWBosonPtr) fWBosonPtr = new GENPDTWBoson();
  for (Int_t m=1; m<=fWBosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fWBosonPtr->GetMode(m); 
     if (mp && (m<fWModesLo || m>fWModesHi)) {
        mp->Lock();
     }
  }
  fWBosonPtr->DebugPrint();

  if (! fZBosonPtr) fZBosonPtr = new GENPDTZBoson();
  fZBosonPtr->DebugPrint();

  if (! fTopPtr) fTopPtr= new GENPDTTQuark(fTopMass,fWBosonPtr);
  fTopPtr->DebugPrint();


  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Double_t rs   = fEcmInit;
  Double_t s    = rs*rs;

  Double_t qw   = kM_w;
  Double_t m0   = kM_e;

  Double_t xilo = TMath::Log((qw-m0)*(qw+m0)/s);
  Double_t xihi = TMath::Log(1.-2.*m0/rs);

  Double_t rxi  = TMath::Exp(xilo);
  Double_t dm0 = (m0*m0/s)*rxi*rxi/(1.-rxi);
  Double_t dp0 = 1.;

  Double_t etlo = -TMath::Log((1.+dm0)/dp0)/2.;
  Double_t ethi =  TMath::Log((1.+dp0)/dm0)/2.;

  Double_t tblo =  qw;
  Double_t tbhi =  fEcmInit;

  Xh_init( 1,     0., fEcmInit*1.1, 50, "Ecm"   );
  Xh_init( 2,   xilo,   xihi,       50, "xi"    );
  Xh_init( 3,   etlo,   ethi,       50, "eta0"  );
  Xh_init( 4,     0.,     1.,       50, "zeta"  );
  Xh_init( 5,     0.,   k2Pi,       50, "phi0"  );
  Xh_init( 6,     0.,   k2Pi,       50, "phi1"  );
  Xh_init( 7,     0.,   k2Pi,       50, "phi10" );
  Xh_init( 8,   tblo,   tbhi,       50, "mtb"   );
  Xh_init( 9,    -1.,    +1.,       50, "costb" );
  Xh_init(10,   -kPi,  +kPi,       50, "phitb" );
  Xh_init(11,   150.,   200.,       50, "mt"    );
  Xh_init(12,    -1.,    +1.,       50, "cost"  );
  Xh_init(13,     0.,   k2Pi,       50, "phit"  );
  Xh_init(14,    55.,   105.,       50, "mw"    );
  Xh_init(15,    -1.,    +1.,       50, "cosf"  );
  Xh_init(16,     0.,   k2Pi,       50, "phif"  );
  Xh_init(17,     0.,     4.,        4, "Helin ");
  Xh_init(18,     0.,     2.,        2, "Helot ");
  Xh_init(19,     0.,    12.,       12, "w mode");
  Xh_init(20,    -1.,    +1.,        2, "CP flg");
  Xh_init(21,    -1.,    +1.,       50, "cos(e)");
  Xh_init(22,     0.,     1.,       50, "xe    ");
  Xh_init(23,    -1.,    +1.,       50, "cos(n)");
  Xh_init(24,     0.,     1.,       50, "ptw/eb");
  Xh_init(25,     0.,     1.,       50, "ee/eb ");
  Xh_init(26,     0.,     1.,       50, "en/eb ");
  Xh_init(27,     0.,     2.,       50, "eW/eb ");
  Xh_init(28,     -s,     0.,       50, "tee   ");
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void ENTBBases::Userout()
{
  cout << "End of ENTBBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcmInit << " [GeV]   "    << endl
       << "Beamstrahlung        = " << (fBeamStr ? "on" : "off")  << endl
       << "Bremsstrahlung       = " << (fISR     ? "on" : "off")  << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"  << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void ENTBBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 4;
   static const Int_t kIHelComb[kNi][2] = {{-1, -1}, // zero
                                           {-1, +1},
                                           {+1, -1}, // zero
                                           {+1, +1}};
   static const Int_t kNf = 2;
                                         //  e  neb bb   b  fu fdb
   static const Int_t kFHelComb[kNf][6] = {{-1, +1, +1, -1, -1, +1},
                                           {+1, +1, +1, -1, -1, +1}};
   Double_t helm = (1. - fPolem)/2.;
   Double_t help = (1. - fPolep)/2.;
   if (fHelCombInitial < helm || helm == 1.) {
      Double_t helcombi = fHelCombInitial/helm;
      if (helcombi < help) fJCombI = 0;
      else                 fJCombI = 1;
   } else {
      Double_t helcombi = (fHelCombInitial-helm)/(1.-helm);
      if (helcombi < help) fJCombI = 2;
      else                 fJCombI = 3;
   }
   if (fJCombI == 0 || fJCombI == 2) {
      weight = 0.;
      return;
   } 
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   fHelFinal  [3] = kFHelComb[fJCombF][3];
   fHelFinal  [4] = kFHelComb[fJCombF][4];
   fHelFinal  [5] = kFHelComb[fJCombF][5];
   weight = kNf;
}

//_______________________________________________________________________
// =====================
//  class GENPDTTQuark
// =====================
//-----------------------------------------------------------------------
GENPDTTQuark::GENPDTTQuark(Double_t m, GENPDTWBoson *wp)
            : GENPDTEntry("t",6,2./3,1./2,m,3,1./2,3*(1+kAlphaS/kPi)),
	      fWBosonPtr(wp)
{
	if (!fWBosonPtr) fWBosonPtr = new GENPDTWBoson();
	Initialize();
}


void GENPDTTQuark::Initialize()
{
   Int_t igu = 2;
         Double_t  spin = 0.5;
         Double_t  cf   = 3*(1. + kAlphaS/kPi);
   for (Int_t igd=0; igd<3; igd++) {
      const Char_t   *namd = kName[1][1][igd];
            Int_t     pidd = kPID [1][1][igd];  
            Double_t  t3d  = -0.5;
            Double_t  qfd  = kChrg[1][1][igd];
            Double_t  md   = kMass[1][1][igd];
            Double_t  vff  = kVkm [igu][igd];
      Double_t gam  = GamToFV(md, kM_w, vff, cf);
      if (gam == 0.) continue;
      GENPDTEntry  *d1p = new GENPDTEntry(namd,-pidd,-qfd, spin, md, igd+1, -t3d, cf);
      GENPDTEntry  *d2p = fWBosonPtr;
      GENDecayMode *dmp = new GENDecayMode(gam);
      dmp->Add(d1p);
      dmp->Add(d2p);
      Add(dmp); 
   }  
}

Double_t GENPDTTQuark::GamToFV(Double_t mf, Double_t mv, Double_t vff, Double_t cf)
{
   Double_t M2  = fMass*fMass;
   Double_t m12 = mf*mf;
   Double_t m22 = mv*mv;
   Double_t x1  = m12/M2;
   Double_t x2  = m22/M2;
   Double_t bet = 1. - 2*(x1+x2) + (x1-x2)*(x1-x2);
   if (bet<0.) return 0.;
            bet = TMath::Sqrt(bet);
   Double_t fac = (kGw*kGw/k4Pi)*vff*vff*(cf/3)/16;
   Double_t gam = fac*TMath::Power(fMass,3)/TMath::Power(mv,2)
	          * bet * ((1-x1)*(1-x1) + x2*(1+x1) - 2*x2*x2);
   return gam;
}
