#ifndef __S4VEXACTHIT__
#define __S4VEXACTHIT__
//*****************************************************************************
//* -----------------
//* class S4VExactHit
//* -----------------
//*
//* (Purpose)
//*   Base class for exact hit objects to provide basic functions to access
//*   exact hit information from Jupiter.
//*
//* (Dependency)
//*   Inherits from:   TObject, TAttLockable
//*   Uses:            TVector3, TVAddress, S4VHit
//*   Provides:        S4VExactHit
//* (Usage)
//*   Override the following in derived classes:
//*       GetPosition(  i):  if pre (i=-1) or post (i=+1) or both are needed
//*       SetPosition(x,i):  
//*
//* (Update Record)
//*   2004/07/30  K.Fujii  Primitive version.
//*
//*****************************************************************************

#include "TAttLockable.h"
#include "TObject.h"
#include "TVector3.h"

class TVAddress;
class S4VHit;


//-----------------------------------------------------------------------------
// -----------------
// Class S4VExactHit
// -----------------
//
#if 0
class S4VExactHit : public TObject, virtual public TAttLockable {
#else
class S4VExactHit : public TObject, public TAttLockable {
#endif
  public:
      enum ESortOption { kHitID = 0, kTrackID, kMotherTrackID, kPDGCode, kCharge,
                         kPosition, kMomentum,  kEnergy, kEnergyDeposit, kTof, 
                         kEventID, kAddress };

  //---------------------------------------------------------------------------
  // =======
  // Methods
  // =======
  //
  // --------------
  // Public Methods
  // --------------
  //
  public:
   //--------------------------------------------------------------------------
   // Ctors and  dtor
   //----------------
   S4VExactHit();
   S4VExactHit(      Int_t       hitid,      // hist ID
                     Int_t       trackid,    // track ID
                     Int_t       momtrackid, // mother track ID
                     Int_t       pid,        // PDG code
                     Double_t    chg,        // charge
               const TVector3  & xv,         // position
                     Double_t    tof,        // TOF
               const TVector3  & pv,         // momentum
                     Double_t    e,          // energy
                     Double_t    edep,       // energy deposit
                     TVAddress * addr,       // address
                     Int_t       evid = 0,   // event ID
                     void      * shtp = 0);  // backward pointer to smeared hit

   virtual ~S4VExactHit() {}

   //--------------------------------------------------------------------------
   // Sorting functions
   //------------------
   virtual Bool_t  IsSortable() const;
   virtual Int_t   Compare   (const TObject * obj) const;

   virtual Int_t   GetSortBy ()                    const { return (Int_t)fSortBy;      }
   static  void    SetSortBy (Int_t opt = kTrackID)      { fSortBy = (ESortOption)opt; }

   //--------------------------------------------------------------------------
   // Getters & setters
   //------------------
   inline virtual Int_t       GetHitID        ()            const  { return fHitID;         }
   inline virtual Int_t       GetTrackID      ()            const  { return fTrackID;       }
   inline virtual Int_t       GetMotherTrackID()            const  { return fMotherTrackID; }
   inline virtual Int_t       GetPDGCode      ()            const  { return fPDGCode;       }
   inline virtual Double_t    GetCharge       ()            const  { return fCharge;        }
   inline virtual TVector3    GetPosition     (Int_t i = 0) const  { return fPosition;      }
   inline virtual TVector3    GetMomentum     ()            const  { return fMomentum;      }
   inline virtual Double_t    GetTotalEnergy  ()            const  { return fEnergy;        }
   inline virtual Double_t    GetEnergyDeposit()            const  { return fEnergyDeposit; }
   inline virtual Double_t    GetTOF          ()            const  { return fTOF;           }
   inline virtual TVAddress * GetAddressPtr   ()            const  { return fAddressPtr;    }
   inline virtual void      * GetSmearedHitPtr()            const  { return fSmearedHitPtr; }
   inline virtual Int_t       GetEventID      ()            const  { return fEventID;       }

   inline virtual void SetHitID        (Int_t      i)              { fHitID         = i;    }
   inline virtual void SetTrackID      (Int_t      i)              { fTrackID       = i;    }
   inline virtual void SetMotherTrackID(Int_t      i)              { fMotherTrackID = i;    }
   inline virtual void SetPDGCode      (Int_t      i)              { fPDGCode       = i;    }
   inline virtual void SetCharge       (Double_t   c)              { fCharge        = c;    }
   inline virtual void SetPosition     (TVector3   x, Int_t i = 0) { fPosition      = x;    }
   inline virtual void SetMomentum     (TVector3   p)              { fMomentum      = p;    }
   inline virtual void SetTotalEnergy  (Double_t   e)              { fEnergy        = e;    }
   inline virtual void SetEnergyDeposit(Double_t   e)              { fEnergyDeposit = e;    }
   inline virtual void SetTOF          (Double_t   t)              { fTOF           = t;    }
   inline virtual void SetAddressPtr   (TVAddress* a)              { fAddressPtr    = a;    }
   inline virtual void SetSmearedHitPtr(void     * p)              { fSmearedHitPtr = p;    }
   inline virtual void SetEventID      (Int_t      i)              { fEventID       = i;    }

   //--------------------------------------------------------------------------
   // Utility functions
   //------------------
  
   virtual void    DebugPrint(Char_t * opt = "") {}

  //---------------------------------------------------------------------------
  // -----------------------
  // Private methods, if any
  // -----------------------
  //
  private:

  //---------------------------------------------------------------------------
  // ============
  // Data Members
  // ============
  //
  private:
           Int_t        fHitID;               //  hit ID
           Int_t        fTrackID;             //  track ID
           Int_t        fMotherTrackID;       //  mother track ID
           Int_t        fPDGCode;             //  PDG code
           Double_t     fCharge;              //  charge
           TVector3     fPosition;            //  position
           Double_t     fTOF;                 //  tof
           TVector3     fMomentum;            //  momentum
           Double_t     fEnergy;              //  total energy
           Double_t     fEnergyDeposit;       //  energy deposit
           TVAddress   *fAddressPtr;          //! pointer to address object
           void        *fSmearedHitPtr;       //! backward pointer to smeared hit
           Int_t        fEventID;             //  event type, buch id??

    static ESortOption  fSortBy;              //! sorting option

  ClassDef(S4VExactHit, 1) // base class for an exact hit
};

//-----------------------------------------------------------------------------
// inline functions, if any
//-----------------------------------------------------------------------------

#endif
