//*************************************************************************
//* ---------------------
//* U4BabyHitWire Class
//* ---------------------
//* (Description)
//*	Class for a single Baby hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	U4BabyHitWire		: Array containing "U4BabyHit"s
//* (Update Record)
//*	2002/08/28  R.Kuboshima
//*************************************************************************

#include <iostream>
#include <iomanip>
#include <math.h>
#include "TRandom.h"
#include "S4BabyHitWire.h"
#include "S4BabyHit.h"
#include "S4BabyExactHit.h"

#define __TMP__
#define __TWOHITSEP__ 0.2
#define __TWOHITSEPZ__ 9999999999.

Double_t S4BabyHitWire::fgTwoHitSeparation  = __TWOHITSEP__;
Double_t S4BabyHitWire::fgTwoHitSeparationZ = __TWOHITSEPZ__;
Bool_t	 S4BabyHitWire::fgTPCMode = kFALSE;	

Double_t S4BabyHitWire::f2Tkpar[] = { -0.665417, 14.0052 };

ClassImp(S4BabyHitWire)

//=====================================================================
//* constructor -------------------------------------------------------

S4BabyHitWire::S4BabyHitWire(U4CDCTimeToXRelation *calib) 
	     : U4CDCHitWire(calib), fKilled(kFALSE)
{
}

//=====================================================================
//* destructor --------------------------------------------------------

S4BabyHitWire::~S4BabyHitWire()
{
	Delete();
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4BabyHitWire::DebugPrint(const Char_t *opt)
{
	std::cerr.setf(std::ios::right);
	std::cerr.setf(std::ios::fixed);
	std::cerr	<< "wire = (" << std::setw(1) << fLayerNumber
		<< "," << std::setw(1) << fCellNumber
		<< "," << std::setw(1) << fWireNumber << ")";
	TIter next(this);
	S4BabyHit *hp;
	Int_t n = 0;
	while ((hp = (S4BabyHit *)next())) {
		if (n != 0) std::cerr << "                  ";
		std::cerr << std::setw(3) << ++n << " ";
		hp->DebugPrint(opt);
	}
	std::cerr.unsetf(std::ios::right);
	std::cerr.unsetf(std::ios::fixed);
}

//=====================================================================
//* Append ------------------------------------------------------------

Int_t S4BabyHitWire::Append(S4BabyHit *ht)
{
	Int_t 	 layer, cell, wire;
	Double_t xwire, ywire;
	ChannelToWire(ht,layer,cell,wire,xwire,ywire);
	
	// Return value = -1:	1st hit.
	//		=  0:	successfully appended.
	//		=  1:	new wire.
	//		=  2:	new cell.
	//		=  3:	new layer.
	
	if (GetEntries() == 0) {
		fLayerNumber  = layer;
		fCellNumber       = cell;
		fWireNumber   = wire;
		fXwire	  = xwire;
		fYwire	  = ywire;
		ht->fWirePtr = this;
		ht->Update();
                Add(ht);
                return -1;
	} else {
		if (layer == fLayerNumber) {
			if (cell == fCellNumber) {
                    		if (wire == fWireNumber) {
                    			ht->fWirePtr = this;
  					ht->Update();
                      			Add(ht);
                        		return 0;
                        	} else {
                        		return 1;
                        	}
                        } else {
                        	return 2;
                        }
                } else {
                        return 3;
                }
        }
}

//=====================================================================
//* ChannelToWire -----------------------------------------------------

void S4BabyHitWire::ChannelToWire(S4BabyHit *hit,
  			         Int_t &layer, Int_t &cell, Int_t &wire, 
  			         Double_t &x, Double_t &y)
{
	S4BabyExactHit* exacthitp = hit->GetExactHitPtr();
	layer = exacthitp->GetLayerNumber();
	cell  = exacthitp->GetCellNumber();
	wire  = exacthitp->GetWireNumber();
	x     = exacthitp->GetWireX();
	y     = exacthitp->GetWireY();
	
#if 0
	std::cerr << "S4BabyHitWire::ChannelToWire: layer = "
	     << layer << " cell = " << cell << " wire = "
	     << wire << " x = " << x << " y = " << y << std::endl;	
#endif
}

//=====================================================================
//* KillHits ----------------------------------------------------------

void S4BabyHitWire::KillHits( ) 
{
	UnSort();
	
	if (!fgTPCMode) {
		S4BabyHit::SetSortBy("TrueDriftLength");
		Sort();
		Double_t dllast = -9999999.;
		Double_t zlast  = -9999999.;
		S4BabyHit *lasthitp = 0;
		Double_t distance  = -9999999.;
		Double_t zdistance = -9999999.;
	
		TIter next(this);
		S4BabyHit *hp;
		while ((hp = (S4BabyHit *)next())) {
			hp->SetLastHitPtr(lasthitp);
			Double_t dlcur = TMath::Abs(hp->GetExactHitPtr()->GetDriftLen());
			distance = dlcur - dllast;
			hp->SetDistanceToPrevious( distance );

			Double_t zcur  = (hp->GetExactHitPtr()->GetPosition()).Z();
			zdistance= TMath::Abs(zcur - zlast);		

#if 1
//-------------------------------------------------------------------------------------------
//  TEMPORARY TREATMENT: Need to check if the last hit is really a delta ray.
//------------------------------------------------------------------------------------------
#if 1
			if (lasthitp && 
			    lasthitp->GetExactHitPtr()->GetPDGCode() == 11 &&
			    lasthitp->GetExactHitPtr()->GetMotherTrackID() == hp->GetTrackID()) {
				hp->SetBit(kEModifiedByDeltaRay);
				lasthitp->SetBit(kECannotSeparate);
				TVector3 pos = lasthitp->GetExactHitPtr()->GetPostPosition();
			        distance = dlcur - pos.x();
			        hp->SetDistanceToPrevious( distance );
			}
#endif
//------------------------------------------------------------------------------------------
//	        	std::cerr << " S4BabyHit::distance = " << distance << std::endl;
	        	Double_t eff = (TMath::TanH((distance-f2Tkpar[0])*f2Tkpar[1])
					   -TMath::TanH(-f2Tkpar[0]*f2Tkpar[1]))
				           /(1.-TMath::TanH(-f2Tkpar[0]*f2Tkpar[1]));
//			std::cerr << " S4BabyHit::eff  = " << eff << std::endl;
			Double_t rand = gRandom->Uniform(0.,1.);
//			std::cerr << " S4BabyHit::rand = " << rand << std::endl;
			if ( rand > eff )
#else
			if ( distance < fgTwoHitSeparation 
					&& zdistance < fgTwoHitSeparationZ ) 
#endif
			{
#if 0
			        std::cerr << " S4BabyHit::Kill:distance = " 
					  << distance << std::endl;
				std::cerr << " S4BabyHit::Kill:Trackno =  " 
					  << hp->GetTrackID() << std::endl;
				std::cerr << " S4BabyHit::Kill:Wireno =  " 
					  << hp->GetWireNumber() << std::endl;
				std::cerr << " S4BabyHit::Kill:Cellno =  " 
					  << hp->GetCellNumber() << std::endl;
#endif
				hp->SetBit(kECannotSeparate);
			}

//-------------------------------------------------------------------------------------------
//  TEMPORARY TREATMENT: Need to check if the last hit is really a delta ray.
//------------------------------------------------------------------------------------------
			if (!lasthitp 
			    || (lasthitp && lasthitp->GetExactHitPtr()
				    ->GetMotherTrackID() != hp->GetTrackID())) {
			        dllast = dlcur;
			        zlast  = zcur;
			}
//------------------------------------------------------------------------------------------
			lasthitp = hp;
		}
	} else {
		S4BabyHit::SetSortBy("ZCoodinate");      // wrong sort ( no central membrane )
		Sort();
		Double_t dllast = -9999999.;
		Double_t zlast  = -9999999.;
		Double_t distance  = -9999999.;
		Double_t zdistance = -9999999.;
	
		TIter next(this);
		S4BabyHit *hp;
		while ((hp = (S4BabyHit *)next())) {
			Double_t zcur = (hp->GetExactHitPtr()->GetPosition()).Z();
			zdistance= zcur - zlast;		
			hp->SetDistanceToPrevious( distance );

			Double_t dlcur  = hp->GetExactHitPtr()->GetDriftLen();
			distance = TMath::Abs(dlcur - dllast);
			if ( zdistance < fgTwoHitSeparationZ && distance < fgTwoHitSeparation ) 
			{
				hp->SetBit(kECannotSeparate);
			}
			dllast = dlcur;	
			zlast  = zcur;	
		}
	}	

	fKilled = kTRUE;
}	
	
		


