#ifndef __S4CALSMALLCLUSTER__
#define __S4CALSMALLCLUSTER__

//*************************************************************************
//* ---------------------
//* S4CALSmallCluster Class
//* ---------------------
//* (Description)
//*	represent calorimeter cluster 
//* (Requires)
//*	S4CALHit
//* (Provides)
//*	S4CALSmallCluster		: made object class
//* (Update Record)
//*	2004/09/03  S.Yamamoto	Original version.
//*	2004/11/04  K.Fujii	Added drawable attribute.
//*************************************************************************

#include "S4VCluster.h"

class S4CALSmallCluster : public S4VCluster{

friend class S4CALSmallClusterMakerBuf;

public:
  S4CALSmallCluster();
  ~S4CALSmallCluster();
  
  void DebugPrint(const Char_t *opt = "");

  //====================================
  //* Getter & Setter
  //------------------------------------

  TVector3    GetILPosition()   { return fILPos;      }
  S4HYBTrack* GetILTrackPtr()   { return fILTrackPtr; }

  Bool_t   IsMIP()          { return fIsMIP;   }
  Bool_t   IsEM()           { return fIsEM;    }

  //====================================================================
  //* PFOMap
  //
  // The small cluster has multimap which discribes "Matched information".
  // (It is set by S4CALRealClusterMakerBuf::MakeClusters())
  // "Matched information" is, for example, "small cluster is
  // gamma/electron-like or not", "small cluster is matched to the track" etc.
  // MatchedType_t discribe the "used method" such as "kgGammaFinder" etc.
  // (MatchedType_t is defined in "../hitmaker/S4CALTypeDef.h")
  // and second component of multimap is some parameters which discribes
  // some kind of "correctness" (chi square, confidence level etc.).
  // These information are used when one try to make "real" clusters.
  //
  // kgEMFitter     : vec[0] = E0(GeV), vec[1] = a, vec[2] = b,
  //                  vec[3] = chi2/ndf, vec[4] = confidence level,
  //                  vec[5] = mother small cluster ID
  // kgEHCompare    : vec[0] = mother small cluster ID, vec[1] = H/E ratio
  // kgTrackMinLayer: vec[0] = track ID, vec[1] = distance from track (cm),
  //                  vec[2] = cut value for distance (cm)
  // kgTrackTube    : vec[0] = track ID, vec[1] = distance from track (cm),
  //                  vec[2] = momentum (GeV/c),
  //                  vec[3] = cut value for radius (cm),
  //                  vec[4] = slope angle of tube (rad)
  // kgTrackThrust  : vec[0] = track ID, vec[1] = track matching (cm),
  //                  vec[2] = cut value (cm)
  // kgILFinder     : vec[0] = track ID, vec[1] = X value of IL position (cm),
  //                  vec[2] = Y value of IL position (cm),
  //                  vec[3] = Z value of IL position (cm)
  // kgEHCompare    : vec[0] = matched pair ID, vec[1] = distance (cm)
  // kgEECopmare    : vec[0] = mother small cluster ID,
  //                  vec[1] = energy ratio (upper/(upper+lower))
  //

  std::multimap<MatchedType_t, std::vector<Double_t> > &GetPFOMap() { return fPFOMap; }     
  void AddPFOMap(MatchedType_t pfofunc, std::vector<Double_t> pfoqual) { fPFOMap.insert(std::pair<MatchedType_t, std::vector<Double_t> >(pfofunc, pfoqual)); }


  //===============================
  //* Cut value getters
  //-------------------------------
  static Int_t       GetMaxNHitsCut()          { return fgMaxNHits;       }
  static Double_t    GetEMEHitCut()            { return fgEMHitEmin;      }
  static Double_t    GetEMTOFHitCut()          { return fgEMHitTOFmax;    } 
  static Double_t    GetHDEHitCut()            { return fgHDHitEmin;      }
  static Double_t    GetHDTOFHitCut()          { return fgHDHitTOFmax;    } 
  static Double_t    GetMaxEnergyCut()         { return fgMaxEnergy;      }

  //===============================
  //* Service method
  //-------------------------------
 public:
  void Add(const S4CALHit* hit);     // Add hit
  void Add(S4CALSmallCluster* scl);  // Add small cluster

  //=================================================================
  //* Data member
  //-----------------------------------------------------------------

 private:
  Bool_t      fIsMIP;          // MIP or not
  Bool_t      fIsEM;           // CalType
  TVector3    fILPos;          // IL position
  S4HYBTrack* fILTrackPtr;     // Hybrid track pointer 

  std::multimap<MatchedType_t, std::vector<Double_t> > fPFOMap;

  //==============================================
  //* Cut values (checked in S4CALSmallClusterMakerBuf)
  //----------------------------------------------
  static Int_t    fgMaxNHits;          // max nhits for a small cluster
  static Double_t fgEMHitEmin;         // Hit energy cut 
  static Double_t fgEMHitTOFmax;       // Hit tof cut 
  static Double_t fgHDHitEmin;         // Hit energy cut 
  static Double_t fgHDHitTOFmax;       // Hit tof cut 
  static Double_t fgMaxEnergy;         // max energy for a small cluster (GeV)

  ////// Cheating method!!! /////////////
 public:
  Bool_t IsMuon(Double_t cut = 0.9);
  Bool_t IsKPI(Double_t cut = 0.9);
  ///////////////////////////////////////

  ClassDef(S4CALSmallCluster, 1) // Cluster data information
};
    
#endif
