#ifndef __S4VCLUSTER__
#define __S4VCLUSTER__

//*************************************************************************
//* ---------------------
//* S4VCluster Class
//* ---------------------
//* (Description)
//*	represent calorimeter cluster 
//* (Requires)
//*	S4CALHit
//* (Provides)
//*	S4VCluster		: made object class
//* (Update Record)
//*	2004/09/03  S.Yamamoto	Original version.
//*	2004/11/04  K.Fujii	Added drawable attribute.
//*************************************************************************

#include "TObjArray.h"
#include "TVector3.h"
#include "S4CALHit.h"
#include "S4CALTypeDef.h"
#include "TAttDrawable.h"
#include "S4HYBTrack.h"

#include <map>
#include <vector>
#include <utility>

class S4VCluster : public TObjArray, public TAttLockable, public TAttDrawable {

public:
  S4VCluster();
  virtual ~S4VCluster();
  
  virtual void 	DebugPrint(const Char_t *opt = "");

  using TAttDrawable::Draw;
  virtual void Draw(Int_t color, const Char_t *opt = "");
 
  //====================================
  //* Getter & Setter
  //------------------------------------

  virtual Int_t      GetCALType()     { if(!fDone) Update(); return fCALType; }
  virtual Double_t   GetEnergy()      { if(!fDone) Update(); return fEnergy;  }
  virtual Double_t   GetEnergyDeposit(){ if(!fDone) Update(); return fEnergyDeposit;}
  virtual TVector3   GetMomentum()    { if(!fDone) Update(); return fMomentum;}
  virtual TVector3   GetPosition()    { if(!fDone) Update(); return fPosition;}
  virtual Double_t   GetTOF()         { if(!fDone) Update(); return fTOF;     }

  virtual TVector3   GetThrustVector(){ if(!fDone) Update(); return fThrustVector;}
  virtual Double_t   GetThrust()      { if(!fDone) Update(); return fThrust;  }
  virtual Double_t   GetBroadening()  { if(!fDone) Update(); return fBroadening;}
  virtual Int_t      GetNHits()       { if(!fDone) Update(); return fNHits;   }
  virtual Double_t   GetMeanLayerID() { if(!fDone) Update(); return fMeanLayerID;}
  virtual Int_t      GetMinLayerID()  { if(!fDone) Update(); return fMinLayerID;}
  virtual Int_t      GetMaxLayerID()  { if(!fDone) Update(); return fMaxLayerID;}
  virtual Int_t      GetClusterID()   { if(!fDone) Update(); return fClusterID;}
  virtual TVector3 GetMinLayerPos()   { if(!fDone) Update(); return fMinPos;}
  virtual TVector3 GetMaxLayerPos()   { if(!fDone) Update(); return fMaxPos;}

  virtual std::vector<S4CALHit*> GetCALHitsVector(){if(!fDone) Update(); return fCALHitsVector;}

  virtual void     SetClusterID(Int_t id) { fClusterID = id; }

  //===============================
  //* Service method
  //-------------------------------
 public:
  virtual void      Add(const S4CALHit *hit);     // Add hit
  virtual void      Update();       // Update cluster information

 protected: 
  virtual void      Calibrate();    // energy calibration
  virtual void      MakeThrust();


  //=================================================================
  //* Data member
  //-----------------------------------------------------------------

 protected:
  Int_t       fCALType;        // CalType
  Double_t    fEnergy;         // Calibrated Energy Deposit of Cluster(GeV)
  Double_t    fEnergyDeposit;  // Energy Deposit of Cluster(GeV)
  TVector3    fPosition;       // Cluster Position x(cm),y(cm),z(cm)
  TVector3    fMomentum;       // Momentum of Cluster(GeV)
  TVector3    fThrustVector;   // unit vector along the thrust axis
  Double_t    fThrust;         // thrust
  Double_t    fBroadening;     // broadening
  Double_t    fTOF;            // hit time at Cell(nsec)
  Int_t       fNHits;          // number of hits in a small cluster
  Bool_t      fDone;           // (1,0) = (small clustering done, not yet)
  Double_t    fMeanLayerID;    // mean layer ID
  Int_t       fMinLayerID;     // minimum layer ID
  Int_t       fMaxLayerID;     // maximum layer ID
  Int_t       fClusterID;      // cluster ID
  std::vector<S4CALHit*> fCALHitsVector; // contained CAL hits vector
  TVector3    fMinPos;         // minimum layer position
  TVector3    fMaxPos;         // maximum layer position

  //==============================================
  //*  temporary calibration factor
 public:
  static Double_t fgCalibFactorForEM;
  static Double_t fgCalibFactorForHD;
  


  ////// Cheating method!!! /////////////
 public:
  virtual std::map<Int_t, Double_t> &GetPIDMap(){ if(!fDone) Update(); return fPIDMap;}
  virtual std::map<Int_t, std::pair<Int_t,Double_t> > &GetPostHitIDMap(){ if(!fDone) Update(); return fPostHitIDMap;}

 protected:
  virtual void SetPIDMap();

  std::map<Int_t, Double_t> fPIDMap;  // <PID,e-dep>
  std::map<Int_t, std::pair<Int_t,Double_t> > fPostHitIDMap;  // <PostHitID,<PID,e-dep>>
  ///////////////////////////////////////


  ClassDef(S4VCluster, 1) // Cluster data information
};
    
#endif
