//*************************************************************************
//* -----------
//* S4JetAnalysis
//* -----------
//* (Description)
//* 	Clasess to analyze Jet data.
//* (Update Record)
//*	2004/11/03  K.Fujii	Original version.
//*	2004/12/08  S.Yamamoto	WW version.
//*************************************************************************

#include "JSFSteer.h"
#include "S4JetAnalysis.h"
#include <iostream>
#include <sstream>

#include "S4GENMCTruthBuf.h"
#include "S4PFObject.h"
#include "S4PFOMakerBuf.h"
#include "S4JetMakerBuf.h"
#include "S4JetSet.h"
#include "S4Jet.h"

#include "ANLPairCombiner.h"

ClassImp(S4JetAnalysis)

//___________________________________________________________________________
S4JetAnalysis::S4JetAnalysis(const Char_t *name, const Char_t *title)
             : JSFModule(name, title)
{
   //std::cerr << " constructor for S4JetAnalysis is called..." << std::endl;
}

//___________________________________________________________________________
S4JetAnalysis::~S4JetAnalysis()
{
}

//___________________________________________________________________________
Bool_t S4JetAnalysis::Initialize()
{
   TDirectory *last = gDirectory;
   gFile->cd("/");

   last->cd();

   return kTRUE;
}


//___________________________________________________________________________
Bool_t S4JetAnalysis::Process(Int_t ev)
{
   Monitor4Jets(ev);
   Monitor2Jets(ev);
#if 1
   std::cerr << "S4JetAnalysis::Process evt " << ev << " is finished" 
             << std::endl;
#endif
   return kTRUE;   
}   

static const Double_t kMassW   =  80.4; // W mass by s.y
static const Double_t kSigmaMW =   4.0; // W mass resolution

static const Double_t xEvis    = 220.0; // Evis cut [GeV}
static const Double_t xPt      = 999.0; // Pt   cut [GeV}
static const Double_t xPl      = 999.0; // Pl   cut [GeV]
static const Double_t xM2j     =  40.0; // mass cut [GeV]
static const Double_t xCosj    =   0.8; // cos(theta_J) cut
static const Int_t    xNpsj    =    10; // cut on #particles/jet

static const Double_t xConeAng = 20.;   // cone angle [degrees]
static const Double_t xElep    = 20.;   // Elep  cut [GeV]
static const Double_t xEcone   = 15.;   // Econe cut [GeV]
static const Double_t xCosCone = TMath::Cos(xConeAng*TMath::Pi()/180.);

static       Int_t    ngood    =     0;
static       Int_t    ngood2   =     0;

//___________________________________________________________________________
void S4JetAnalysis::Monitor4Jets(Int_t ev)
{
   using namespace std;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   // Prepare cut table.
   static TH1D *hStat = 0;
   if (!hStat) hStat = new TH1D("hStat", "Cut Table", 20, 0, 20);
   Int_t cutid = 0;
   hStat->Fill(cutid++,1.);

   // Get Jet buffer
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>(
                            gJSF->FindModule("S4JetMaker")->EventBuf()
                         );

   // Get jets with default ycut.
   Int_t njets = 0;
   S4JetSet      *jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat->Fill(cutid++,1.);

   ANL4DVector qsum;
   TIter next(jsp);
   S4Jet *jp = 0;
   while ((jp = static_cast<S4Jet *>(next()))) qsum += *jp;

   Double_t evis = qsum(0);      // Evis
   Double_t pt   = qsum.GetPt(); // P_t
   Double_t pl   = qsum(3);      // P_l

   // Evis, Pt, and Pl cuts.
   static TH1D *hEvis = 0;
   if (!hEvis) hEvis = new TH1D("hEvis", "Evis", 500, 0, 500);
   hEvis->Fill(evis,1.);
   if (evis < xEvis) return;
   hStat->Fill(cutid++,1.);

   static TH2D *hPtPl = 0;
   if (!hPtPl) hPtPl = new TH2D("hPtPl", "(Pt,Pl)", 100,   0., 100.,
                                                    100, -50.,  50.);
   hPtPl->Fill(pt,pl,1.);
   if (pt > xPt || TMath::Abs(pl) > xPl) return;
   hStat->Fill(cutid++,1.);

   // Select 4 jets.
   njets = 4;
   jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat->Fill(cutid++,1.);

   static TH1D *hEjet = 0;
   if (!hEjet) hEjet = new TH1D("hEjet", "Ejet for 4 Jets", 200, 0, 200);

   TIter nextjet(jsp);
   while ((jp = static_cast<S4Jet *>(nextjet()))) hEjet->Fill(jp->E(),1.);

   static TH2D *hMwMw = 0;
   if (!hMwMw) hMwMw = new TH2D("hMwMw", "(Mw,Mw) for 4 Jets", 200,  0., 200.,
                                                               200,  0., 200.);

   // Mass cuts to select WW candidates.
   S4JetSet &jets = *jsp;
   TObjArray solutions(10);
   solutions.SetOwner();
   ANLPairCombiner w1candidates(jets,jets);
   ANLPair *w1p;
   while ( (w1p = (ANLPair*)w1candidates()) ) {
      ANLPair &w1 = *w1p;
      Double_t w1mass = w1().GetMass();
      if (TMath::Abs(w1mass - kMassW) > xM2j) continue;
      w1.LockChildren();
      ANLPairCombiner w2candidates(w1candidates);
      ANLPair *w2p;
      while ( (w2p = (ANLPair*)w2candidates()) ) {
         ANLPair &w2 = *w2p;
         if (w2.IsLocked()) continue;
         Double_t w2mass = w2().GetMass();
         if (TMath::Abs(w2mass - kMassW) > xM2j) continue;
         hMwMw->Fill(w1mass, w2mass, 1.);
#if 1
         cerr << " M_W1  = " << w1mass << " M_W2 = " << w2mass << endl;
         cerr << " w1p   = " << (void*)w1p
              << " w2p   = " << (void*)w2p << endl;
         cerr << " w1[0] = " << (void*)w1[0]
              << " w1[1] = " << (void*)w1[1]
              << " w2[0] = " << (void*)w2[0]
              << " w2[1] = " << (void*)w2[1] << endl;
#endif
         Double_t chi2 = TMath::Power((w1mass - kMassW)/kSigmaMW,2.)
                       + TMath::Power((w2mass - kMassW)/kSigmaMW,2.);
         solutions.Add(new ANLPair(w1p,w2p,chi2));
      }
      w1.UnlockChildren();
   }

   if (!solutions.GetEntries()) return;
   hStat->Fill(cutid++,1.);

   // Found some solutions.
   // Do further analysis here.

   solutions.Sort();

   static TH2D *hMwMw4 = 0;
   if (!hMwMw4) hMwMw4 = new TH2D("hMwMw4", "(Mw,Mw) for 4 Jets", 200,  0., 200.,
                                                               200,  0., 200.);

   TIter nextsol(&solutions);
   ANLPair &sol = *static_cast<ANLPair *>(nextsol());
   ANLPair &w1  = *static_cast<ANLPair *>(sol[0]);
   ANLPair &w2  = *static_cast<ANLPair *>(sol[1]);
   hMwMw4->Fill(w1.GetMass(), w2.GetMass(), 1.);

   // Count up good events.
   ngood++;
   // That's it, folks!
   last->cd();
}

//___________________________________________________________________________
void S4JetAnalysis::Monitor2Jets(Int_t ev)
{
   using namespace std;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   // Prepare cut table.
   static TH1D *hStat2 = 0;
   if (!hStat2) hStat2 = new TH1D("hStat2", "Cut Table for 2 Jets", 40, 0, 20);
   Int_t cutid = 0;
   hStat2->Fill(cutid++,1.); // cut 1

   // Get Jet buffer
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>(
                            gJSF->FindModule("S4JetMaker")->EventBuf()
                         );

   // Look for isolated leptons and lock it.
   S4PFOMakerBuf *pmbp = static_cast<S4PFOMakerBuf *>(
                            gJSF->FindModule("S4PFOMaker")->EventBuf()
                         );
   TObjArray &pfos = *pmbp->GetOutputBufPtr();
   TObjArray ileps;
   TIter nextpfo(&pfos); // prepare cluster iterator
   S4PFObject *pfop = 0;
   while ((pfop = static_cast<S4PFObject *>(nextpfo()))) {
      Double_t econe = pfop->GetConeEnergy(xCosCone, &pfos);
      Double_t elep  = pfop->E();
      if (elep > xElep && econe < xEcone) {
         ileps.Add(pfop);
         pfop->Lock();
      }
   }

   // Get jets with default ycut.
   Int_t njets = 0;
   S4JetSet      *jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.); // cut 2

   ANL4DVector qsum;
   TIter next(jsp);
   S4Jet *jp = 0;
   while ((jp = static_cast<S4Jet *>(next()))) qsum += *jp;

   Double_t evis = qsum(0);      // Evis
   Double_t pt   = qsum.GetPt(); // P_t
   Double_t pl   = qsum(3);      // P_l

   // Evis, Pt, and Pl cuts.
   static TH1D *hEvis2 = 0;
   if (!hEvis2) hEvis2 = new TH1D("hEvis2", "Evis for 2 Jets", 500, 0, 500);
   hEvis2->Fill(evis,1.);
   if (evis > xEvis) return;
   hStat2->Fill(cutid++,1.); // cut 3

   static TH2D *hPtPl2 = 0;
   if (!hPtPl2) hPtPl2 = new TH2D("hPtPl2", "(Pt,Pl) for 2 Jets", 
                                          100,   0., 100., 100, -50.,  50.);
   hPtPl2->Fill(pt,pl,1.);
   if (pt > xPt || TMath::Abs(pl) > xPl) return;
   hStat2->Fill(cutid++,1.); // cut 4
    
   // Select 2 jets.
   njets = 2;
   jsp  = jmbp->GetJetSetPtr(njets);
   if (!jsp) return;
   hStat2->Fill(cutid++,1.); // cut 5

   static TH1D *hEjet2 = 0;
   if (!hEjet2) hEjet2 = new TH1D("hEjet2", "Ejet for 2 Jets", 200, 0, 200);

   static TH1D *hCosj2 = 0;
   if (!hCosj2) hCosj2 = new TH1D("hCosj2", "Cosj for 2 Jets", 200, -1., 1.);

   static TH1D *hNpsj2 = 0;
   if (!hNpsj2) hNpsj2 = new TH1D("hNpsj2", "Np/Jet for 2 Jets", 100, 0, 100);

   static TH1D *hElep2 = 0;
   if (!hElep2) hElep2 = new TH1D("hElep2", "Elep for 2 Jets", 200, 0, 200);

   static TH1D *hCosl2 = 0;
   if (!hCosl2) hCosl2 = new TH1D("hCoslep2", "Coslep for 2 Jets", 200, -1., 1.);

   Int_t njok = 0;
   TObjArray jets;
   TIter nextjet(jsp);
   while ((jp = static_cast<S4Jet *>(nextjet()))) {
     if (jp->GetNparticles() == 1) {
       const TObjArray &trks = jp->GetParticlesInJet();
       S4PFObject *lp = static_cast<S4PFObject *>(trks.At(0));
       if (lp->GetPID() == PID::kgElectron_id 
           || lp->GetPID() == PID::kgMuon_id ) {
	 hElep2->Fill(lp->E(),1.);
	 hCosl2->Fill(lp->CosTheta(),1.);
	 continue;
       }
     }
      hEjet2->Fill(jp->E(),1.);
      hCosj2->Fill(jp->CosTheta(),1.);
      hNpsj2->Fill(jp->GetNparticles(),1.);
#if 1
      if (TMath::Abs(jp->CosTheta()) < xCosj 
          && jp->GetNparticles() > xNpsj) {
#endif
	njok++;
	jets.Add(jp);
#if 1 
     }
#endif
   }
   if (njok < 2) return;
   hStat2->Fill(cutid++,1.); // cut 6


   static TH2D *hMwMl2 = 0;
   if (!hMwMl2) hMwMl2 = new TH2D("hMwMl2", "(Mw,Ml) for 2 Jets",
                                            200,  0., 200., 100,  0., 400.);

   // Mass cuts to select Wlnu candidates.
   
   ANLPairCombiner wcandidates(jets,jets);
   ANLPair &w = *static_cast<ANLPair *>(wcandidates());

   S4GENMCTruthBuf *genbp = static_cast<S4GENMCTruthBuf *>(
                              gJSF->FindModule("S4GENMCTruth")->EventBuf()
                           );
   Double_t Ecm = genbp->GetPVcm().E();

   ANL4DVector qcm(Ecm);
   Double_t wjmass = w().GetMass();
   Double_t wlmass = (qcm - w).GetMass();
#if 1
   cerr << " M_W = " << wjmass << " M_lnu = " << wlmass << endl;
#endif
   
   hMwMl2->Fill(wjmass, wlmass, 1.);
#if 1
   if (TMath::Abs(wjmass - kMassW) > xM2j || TMath::Abs(wlmass - kMassW) > xM2j) return;
#endif 
  hStat2->Fill(cutid++,1.); // cut 7

   // Found a solution.
   // Do further analysis here.


   // Count up good events.
   ngood2++;
   // That's it, folks!
   last->cd();
}

//___________________________________________________________________________
#if 0
void S4JetAnalysis::Draw(Option_t/* *opt */)
{
   gFile->cd("/");
   gDirectory->cd();
}
#endif
