//*************************************************************************
//* ---------------------
//* S4TPCHit Class
//* ---------------------
//* (Description)
//*	Class for a single TPC hit.
//* (Requires)
//*	TObject
//* (Provides)
//*	S4TPCHit		: TObject containing single hit data
//* (Update Record)
//*	2004/08/26  A.Yamaguchi	Original version.
//*************************************************************************

#include <iostream>
#include "S4TPCHit.h"
#include "U4TPCHit.h"
#include "TRandom.h"
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,15,0)
#include "TMath.h"
#endif
using namespace std;

ClassImp(S4TPCHit)

Double_t S4TPCHit::fgDZdrift = 0.04;
Double_t S4TPCHit::fgDRPhi   = 0.015;
Double_t S4TPCHit::fgDVdrift = 5.e-3;

Int_t    S4TPCHit::fgS4TPCHitSortBy = 0; 
TString  S4TPCHit::fgSpatialResolutionCurve = "ILD"; // ILD is default. 2008-2-20
Double_t S4TPCHit::fgT0offset = 0.;

Double_t S4TPCHit::fgSigma0 = 5.5e-3;     // [cm]
Double_t S4TPCHit::fgCd     = 5.533e-3;   // [cm/cm^{0.5}]
Double_t S4TPCHit::fgAlpha  = 0.4444;
Double_t S4TPCHit::fgNe     = 63;

Double_t S4TPCHit::fgSig00 = 0.005 ;  //[cm]
Double_t S4TPCHit::fgSig0SinDep = 0.090 ; // [cm]
Double_t S4TPCHit::fgSigLDep = 0.00053 ; //[cm/cm^{0.5}]
Double_t S4TPCHit::fgSigDZLDep = 0.0080 ; // [cm/cm^{0.5}]

//=====================================================================
//* constructor -------------------------------------------------------

S4TPCHit::S4TPCHit(S4TPCExactHit *exacthit)
        : U4TPCHit(exacthit->GetPadRowNumber(),
                   exacthit->GetPadRowRadius(),
                   exacthit->GetSide(),
                   exacthit->GetPhi()),
          fExactHitPtr(exacthit)
{ 
   exacthit->SetSmearedHitPtr(this);
}

//=====================================================================
//* destructor --------------------------------------------------------

S4TPCHit::~S4TPCHit() 
{
}
  
//=====================================================================
//* SetSortBy ------------------------------------------------------=--

void S4TPCHit::SetSortBy(Char_t *opt)
{
   switch (opt[2]) {
      case 'a':    // sort by "TrackID"
         fgS4TPCHitSortBy = 0;
         break;
      case 'm':    // sort by "Time"
         fgS4TPCHitSortBy = 1;
         break;
      case 'e':    // sort by "EnergyDeposit"
         fgS4TPCHitSortBy = 2;
         break;
      case 'i':    // sort by "Phi"
         fgS4TPCHitSortBy = 3;
         break;
      case 'o':    // sort by "ZCoordinate"
         fgS4TPCHitSortBy = 4;
         break;
      case 'y':    // sort by "LayerID"
         fgS4TPCHitSortBy = 5;
         break;
      case 'd':    // sort by "PadRow"
      default:
         fgS4TPCHitSortBy = 6;
   }
}

//=====================================================================
//* Compare -----------------------------------------------------------

Int_t S4TPCHit::Compare(const TObject *obj) const
{
   switch (fgS4TPCHitSortBy) {
      case 0: {
         Int_t trackid  = fExactHitPtr->GetTrackID();
         Int_t trackido = dynamic_cast<const S4TPCHit *>(obj)
                                           ->fExactHitPtr->GetTrackID();
         return (trackid < trackido ? -1 : (trackid > trackido ? 1 : 0));
      }
      case 1: {
         Double_t tof  = fExactHitPtr->GetTOF();
         Double_t tofo = dynamic_cast<const S4TPCHit *>(obj)
                                          ->fExactHitPtr->GetTOF();
         return (tof < tofo ? -1 : (tof > tofo ? 1 : 0));
      }
      case 2: {
         Double_t edep  = fExactHitPtr->GetEnergyDeposit();
         Double_t edepo = dynamic_cast<const S4TPCHit *>(obj)
                                           ->fExactHitPtr->GetEnergyDeposit();
         return (edep < edepo ? -1 : (edep > edepo ? 1 : 0));
      }   
      case 3: {
         Double_t phi  = const_cast<S4TPCHit *>(this)->GetPhi();
         Double_t phio = ((S4TPCHit *)obj)->GetPhi();
       return (phi < phio ? -1 : (phi > phio ? 1 : 0));
      }
      case 4: {
         Double_t zdrift  = const_cast<S4TPCHit *>(this)->GetZdrift();
         Double_t zdrifto = ((S4TPCHit *)obj)->GetZdrift();
         return (zdrift < zdrifto ? -1 : (zdrift > zdrifto ? 1 : 0));
      }
      case 5: {
         Int_t prn  = GetPadRowNumber();
         Int_t prno = dynamic_cast<const S4TPCHit *>(obj)->GetPadRowNumber();
         return (prn < prno ? -1 : (prn > prno ? 1 : 0));
      }
      case 6:
      default: {
         Double_t pr  = GetPadRowRadius();
         Double_t pro = dynamic_cast<const S4TPCHit *>(obj)->GetPadRowRadius();
         return (pr < pro ? -1 : (pr > pro ? 1 : 0));
         break;
      }
   }
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4TPCHit::DebugPrint(const Char_t *opt)
{
   if (!IsDone()) Update();

   cerr << right << fixed
        << "--------------------------------------------------"       << endl
        << " Track ID   = " << fExactHitPtr->GetTrackID()             << endl
        << " Etot       = " << fExactHitPtr->GetTotalEnergy()         << endl
        << " Pad Row No = " << GetPadRowNumber()                      << endl
        << " Side       = " << (GetSide() < 0 ? "-Z" : "+Z")          << endl
        << " Rpadrow    = " << GetPadRowRadius()                      << endl
        << " R * Phi    = " << GetRPhi()   << " +/- " << GetDRPhi()   << endl
        << " Zdrift     = " << GetZdrift() << " +/- " << GetDZdrift() << endl;
}

//=====================================================================
//* Update ------------------------------------------------------------

void S4TPCHit::Update()
{
   U4TPCHit::Update();
}

//=====================================================================
//* CalculatePosition() -----------------------------------------------

void S4TPCHit::CalculatePosition()
{
   if (fgSpatialResolutionCurve == "experimental") {
      // -----------------------------------------------------
      // sigma^2 = sigma0^2 + Cd^2/(alpha*Ne)*Zdrift
      // -----------------------------------------------------
      Double_t sigma0 = fgSigma0;     // [cm]
      Double_t Cd     = fgCd;   // [cm/cm^{0.5}]
      Double_t alpha  = fgAlpha;
      Double_t Ne     = fgNe;
#if 1
      Ne /= TMath::Sin(fExactHitPtr->GetMomentum().Theta());
#endif
      Double_t sigma  = TMath::Sqrt(sigma0*sigma0 + Cd*Cd/alpha/Ne
                                    * fExactHitPtr->GetZdrift());
      SetDRPhi(sigma);
      SetDZdrift(fgDZdrift); 
   }
   else if ( fgSpatialResolutionCurve == "ILD" ) {
/*
  Formula is based on the mail from Ron Settles on Jan. 30
1) As a function of drift-distance L_{drift}, the expression for the
 r\phi point resolution is, as you know,

 sigma_{point}2 = sigma_02 + Cd2/N_{eff} * L_{drift}

 Proposal 1) on point resolution:
 -------------------------------
 =>sigma_02 = (50micron)^2 + (900micron*sin(phi))^2
 (where phi is the local azimuthal angle of track wrt the padrow)

 =>Cd^2/N_{eff}}=25^2/(22/sin(theta)*h/6mm)=(5.3micron/sqrt(cm))^2*(6mm/h)*sin(theta)
 (this is for B=4T which we favor, h is the pad height=pad-row pitch in mm,
 theta is the polar angle)

 =>sigma_z = 400micron
 =>sigma_z(z)= sqrt{400micron**2 + z(cm) x (80micron/sgrt{cm})**2}
*/
/*
   Typical input values to get Ron's spetial resolution are,
       fgSig00  = 0.005 ; [cm] 
       fgSig0SinDep = 0.090 ; [cm]
       fgSigLDep = 0.00053 ; [cm]  for 4Tesla, 6mm pad pitch
*/

      TVector3 tmom=fExactHitPtr->GetMomentum();
      TVector3 tpos=fExactHitPtr->GetPosition();
      Double_t sinphi=( tmom.X()*tpos.Y() - tmom.Y()*tpos.X() ) 
	/ ( tmom.Perp() *  tpos.Perp() ) ;

      Double_t sig0sq = fgSig00*fgSig00 + fgSig0SinDep*fgSig0SinDep*sinphi*sinphi;
      Double_t sintheta =tmom.Perp()/tmom.Mag(); 
      Double_t ldepcof = fgSigLDep*fgSigLDep*sintheta;

      Double_t sigma  = TMath::Sqrt(sig0sq + ldepcof * fExactHitPtr->GetZdrift());
      SetDRPhi(sigma);
      Double_t sigz= TMath::Sqrt(fgDZdrift*fgDZdrift 
	             + fgSigDZLDep*fgSigDZLDep*fExactHitPtr->GetZdrift());

      SetDZdrift(sigz); 
   } 
   else {
      SetDRPhi(fgDRPhi); 
      SetDZdrift(fgDZdrift); 
   }
   SetRPhi(GetPadRowRadius() * fExactHitPtr->GetPhi()
         + gRandom->Gaus(0., GetDRPhi()));

   SetVdrift (fgDVdrift); 
   Double_t zdrift = TMath::Abs(fExactHitPtr->GetZdrift()) 
                   + gRandom->Gaus(0., GetDZdrift())
                   + GetVdrift() * fgT0offset;
   SetZdrift(zdrift); 
   SetPhi(GetRPhi() / GetPadRowRadius());
}

// ====================================================
//
void S4TPCHit::Print()
{
   cerr << "S4TPCHit :: Parameters for Hit point making " << endl;    
   cerr << "   Method = " << fgSpatialResolutionCurve << endl;
   if (fgSpatialResolutionCurve == "experimental") {
     cerr << "   sigma0 = " << fgSigma0 << "(cm)" << endl;
     cerr << "   Cd     = " << fgCd << "(cm/cm^{1/2})" << endl;
     cerr << "   alpha  = " << fgAlpha << endl;
     cerr << "   Ne     = " << fgNe << endl;
   }
   else if ( fgSpatialResolutionCurve == "ILD" ) {
     cerr << "   Sig00      = " << fgSig00 << " (cm)" << endl;
     cerr << "   Sig0SinDep = " << fgSig0SinDep << "(cm)" << endl;
     cerr << "   SigLDep    = " << fgSigLDep << "(cm)" << endl;
     cerr << "   SigDZDrift = " << fgDZdrift << "(cm)" << endl;
     cerr << "   SigDZLDep  = " << fgSigDZLDep << "(cm)" << endl;
   }
   else if ( fgSpatialResolutionCurve == "flat " ) {
     cerr << "   SigRPhi    = " << fgDRPhi << "(cm)" << endl;
     cerr << "   SigDZDrift = " << fgDZdrift << "(cm)" << endl;
   }

}
