//*************************************************************************
//* ---------------------
//* U4BabyTrack Class
//* ---------------------
//* (Description)
//*	Class to store "U4CDCVectorHit"s for a single track.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4BabyTrack		: Array containing "U4CDCVectorHit"s
//* (Update Record)
//*	2002/11/15   R.Kuboshima    Original Ver.
//*************************************************************************

#include <iostream>
#include <iomanip>
#include "TMath.h"
#include "U4BabyTrack.h"
#include "U4CDCHitWire.h"
#include "U4CDCHit.h"

#define __TMP__

#ifdef __TMP__
#define __BFIELD__ 30.
#endif

Int_t U4BabyTrack::fWireToLookAt = 2;      // wire to look at
Int_t U4BabyTrack::fCellToLookAt = 0;      // wire to look at

ClassImp(U4BabyTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4BabyTrack::U4BabyTrack()
           : fa(0.), fb(0.), fEaa(0.), fEba(0.), fEbb(0.)
{
   SetOwner();
}

U4BabyTrack::U4BabyTrack(const U4BabyTrack &tk) 
           : U4CDCTrack(tk),
             fa(0.), fb(0.), fEaa(0.), fEba(0.), fEbb(0.)
{
   TIter next(&tk);
   U4CDCVectorHit *vhitp;
   while ((vhitp = (U4CDCVectorHit *)next())) {
      Add(new U4CDCVectorHit(*vhitp));
   }
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4BabyTrack::~U4BabyTrack()
{
}


//=====================================================================
//* Compare -----------------------------------------------------------

Int_t U4BabyTrack::Compare(const TObject *obj) const
{
   if (!IsDone()) ((U4BabyTrack *)this)->Update();
   if (!((U4BabyTrack *)obj)->IsDone()) ((U4BabyTrack *)obj)->Update();
   U4BabyTrack *tko    = (U4BabyTrack *)obj;
   switch (GetSortBy()) {
      case 1: {// sort by time: faster one first:-)
         TIter next(this);
         TIter nexto(tko);
         U4CDCVectorHit *vthp;
         U4CDCVectorHit *vthpo;
         while ((vthp = (U4CDCVectorHit *)next())) {
            while ((vthpo = (U4CDCVectorHit *)nexto())) {
               U4CDCTrackHit  *thp     = (U4CDCTrackHit *)vthp->At(0);
               U4CDCHit       *hp      = (U4CDCHit *)thp->GetHitPtr();
               U4CDCTrackHit  *thpo    = (U4CDCTrackHit *)vthpo->At(0);
               U4CDCHit       *hpo     = (U4CDCHit *)thpo->GetHitPtr();
               Int_t           cell    = hp->GetWirePtr()->GetCellID();
               Int_t           cello   = hpo->GetWirePtr()->GetCellID();
               if (cell != fCellToLookAt || cello != fCellToLookAt) return 0;
               Int_t           nwires  = vthp->GetEntries();
               Double_t        wir1st  = hp->GetWireNumber();
               U4CDCTrackHit  *thplst  = (U4CDCTrackHit *)vthp->At(nwires-1);
               U4CDCHit       *hplst   = (U4CDCHit *)thplst->GetHitPtr();
               Double_t        wirlst  = hplst->GetWireNumber();
               Double_t        xwir1st = hp->Xwire();
               Double_t        xwirlst = hplst->Xwire();
               Double_t        xwire   = xwir1st + (xwirlst-xwir1st)
                                       *(fWireToLookAt-wir1st)/(wirlst-wir1st);
               Double_t        ywir1st = hp->Ywire();
               Double_t        ywirlst = hplst->Ywire();
               Double_t        y       = ywir1st + (ywirlst-ywir1st)
                                       *(fWireToLookAt-wir1st)/(wirlst-wir1st);
               Double_t time  = TMath::Abs(((U4BabyTrack *)this)->Xat(y)-xwire);
               Double_t timeo = TMath::Abs(tko->Xat(y)-xwire);
               if (time < timeo) return -1;
               else if (time > timeo) return 1;
               else return 0;
            }
         }
      }
      case 2: {
	      Int_t track  = GetTrackNumber();
	      Int_t tracko = ((U4BabyTrack *)obj)->GetTrackNumber();
	      if (track < tracko) return -1;
	      else if (track > tracko) return 1;
	      else return 0;
      }
      default: {// sort by chi2: better one first:-)
         Double_t chi2  = ((U4BabyTrack *)this)->GetChi2();
         Double_t chi2o = ((U4BabyTrack *)obj)->GetChi2();
         if      (chi2 < chi2o) return -1;
         else if (chi2 > chi2o) return 1;
         else                   return 0;
      }
   }
}


void U4BabyTrack::SetSortBy(Char_t *opt, Int_t wireno)
{
        switch (opt[0]) {
                case 'T':       // sort by "Time"
                        SetSortBy(1);
                        fWireToLookAt = wireno;
                        break;
		case 'S':       // sort by "SerialNumber"
                        SetSortBy(2);
		        break;
                default:        // sort by "Chi2"
                        SetSortBy(0);
        }
}


//=====================================================================
//* Fit ----------------------------------------------------------
void U4BabyTrack::Fit()
{
        Double_t det, x, y, rs2;
        Double_t syy = 0.;
        Double_t sy  = 0.;
        Double_t s   = 0.;
        Double_t sxy = 0.;
        Double_t sx  = 0.;
        TIter nextvhitp (this);
        U4CDCVectorHit *vhitp;
        U4CDCTrackHit *tkhitp;
        while ((vhitp = (U4CDCVectorHit *)nextvhitp())) {
            TIter nexttkhitp(vhitp);
            while ((tkhitp = (U4CDCTrackHit *)nexttkhitp())) {
                U4CDCTrackHit  &tkhit  = *tkhitp;
                U4CDCHit       &hit   = *tkhit.GetHitPtr();
                if (tkhit.IsLocked()) continue;
                rs2  = 1./TMath::Power(hit.DXdrift(),2);
                x    = hit.Xwire()
                             + tkhit.GetLeftRight() * hit.Xdrift();
                y    = hit.Ywire();
                syy += rs2*y*y;
                sy  += rs2*y;
                s   += rs2;
                sxy += rs2*x*y;
                sx  += rs2*x;
            }
        }
        det  = syy*s -sy*sy;
        fa   = (  s*sxy -  sy*sx)/det;
        fb   = (-sy*sxy + syy*sx)/det;
        fEaa =   s/det;
        fEba = -sy/det;
        fEbb = syy/det;

        Double_t chi2 = 0.;
        nextvhitp.Reset();
        while ((vhitp = (U4CDCVectorHit *)nextvhitp())) {
            TIter nexttkhitp(vhitp);
	    while ((tkhitp = (U4CDCTrackHit *)nexttkhitp())) {
                U4CDCTrackHit  &tkhit = *tkhitp;
                U4CDCHit       &hit   = *tkhit.GetHitPtr();
                x = hit.Xwire()
                          + tkhit.GetLeftRight() * hit.Xdrift();
                y = hit.Ywire();
                tkhit.SetResidual(x - fa*y - fb);
#if 0 
                std::cerr << "U4BabyTrack::Fit: Residual = "
                          << tkhit.GetResidual()
                          << ", DXdrift = " << hit.DXdrift()
			  << ", Chi2 = " << chi2
                          << std::endl;
#endif
                chi2 += TMath::Power(tkhit.GetResidual()/hit.DXdrift(),2);
            }
        }
        SetChi2(chi2);
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4BabyTrack::DebugPrint(const Char_t *opt)
{
        std::cerr << "TrackNumber = " << GetTrackNumber()
                  << "  NEntries = "  << GetEntries() << std::endl;
        if (strstr(opt,"Detailed")) {
           if (!IsDone()) Update();
           std::cerr.setf(std::ios::right);
           std::cerr.setf(std::ios::fixed);
           std::cerr << " Type = " << std::setw(2) << GetTrackType()
                     << " chi2 = " << std::setw(6) << std::setprecision(3) << GetChi2()
                     << " (a,b) = (" << std::setw(6) << std::setprecision(3) << fa
                     << "," << std::setw(6) << std::setprecision(3) << fb << ")"
                     << std::endl;
        }

	TIter next(this);
	U4CDCVectorHit *vectorhit;
	while ((vectorhit = (U4CDCVectorHit *)next())) {
		vectorhit->DebugPrint(opt);
	}
}

